# SPDX-License-Identifier: BSD-2-Clause
# torii: UnusedElaboratable=no

import os
from contextlib    import contextmanager
from warnings      import catch_warnings

from torii.util    import flatten
from torii.hdl.ast import *
from torii.hdl.cd  import *
from torii.hdl.mem import *
from torii.hdl.rec import *
from torii.hdl.dsl import *
from torii.hdl.ir  import *
from torii.sim     import *

from ..utils       import *

class SimulatorUnitTestCase(ToriiTestSuiteCase):
	def assertStatement(self, stmt, inputs, output, reset = 0):
		inputs = [Value.cast(i) for i in inputs]
		output = Value.cast(output)

		isigs = [ Signal(i.shape(), name = n) for i, n in zip(inputs, 'abcd') ]
		osig  = Signal(output.shape(), name = 'y', reset = reset)

		stmt = stmt(osig, *isigs)
		frag = Fragment()
		frag.add_statements(stmt)
		for signal in flatten(s._lhs_signals() for s in Statement.cast(stmt)):
			frag.add_driver(signal)

		sim = Simulator(frag)

		def process():
			for isig, input in zip(isigs, inputs):
				yield isig.eq(input)
			yield Settle()
			self.assertEqual((yield osig), output.value)

		sim.add_process(process)
		with sim.write_vcd('test.vcd', 'test.gtkw', traces = [ *isigs, osig ]):
			sim.run()

	def test_invert(self):
		self.assertStatement(
			lambda y, a: y.eq(~a),
			[Const(0b0000, 4)], Const(0b1111, 4)
		)
		self.assertStatement(
			lambda y, a: y.eq(~a),
			[Const(0b1010, 4)], Const(0b0101, 4)
		)
		self.assertStatement(
			lambda y, a: y.eq(~a),
			[Const(0, 4)], Const(-1, 4)
		)

	def test_neg(self):
		self.assertStatement(
			lambda y, a: y.eq(-a),
			[Const(0b0000, 4)], Const(0b0000, 4)
		)
		self.assertStatement(
			lambda y, a: y.eq(-a),
			[Const(0b0001, 4)], Const(0b1111, 4)
		)
		self.assertStatement(
			lambda y, a: y.eq(-a),
			[Const(0b1010, 4)], Const(0b0110, 4)
		)
		self.assertStatement(
			lambda y, a: y.eq(-a),
			[Const(1, 4)], Const(-1, 4)
		)
		self.assertStatement(
			lambda y, a: y.eq(-a),
			[Const(5, 4)], Const(-5, 4)
		)

	def test_bool(self):
		self.assertStatement(
			lambda y, a: y.eq(a.bool()),
			[Const(0, 4)], Const(0)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.bool()),
			[Const(1, 4)], Const(1)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.bool()),
			[Const(2, 4)], Const(1)
		)

	def test_as_unsigned(self):
		self.assertStatement(
			lambda y, a, b: y.eq(a.as_unsigned() == b),
			[Const(0b01, signed(2)), Const(0b0001, unsigned(4))], Const(1)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a.as_unsigned() == b),
			[Const(0b11, signed(2)), Const(0b0011, unsigned(4))], Const(1)
		)

	def test_as_unsigned_lhs(self):
		self.assertStatement(
			lambda y, a: y.as_unsigned().eq(a),
			[Const(0b01, unsigned(2))], Const(0b0001, signed(4))
		)

	def test_as_signed(self):
		self.assertStatement(
			lambda y, a, b: y.eq(a.as_signed() == b),
			[Const(0b01, unsigned(2)), Const(0b0001, signed(4))], Const(1)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a.as_signed() == b),
			[Const(0b11, unsigned(2)), Const(0b1111, signed(4))], Const(1)
		)

	def test_as_signed_issue_502(self):
		self.assertStatement(
			lambda y, a: y.eq(a.as_signed()),
			[Const(0b01, unsigned(2))], Const(0b0001, signed(4))
		)
		self.assertStatement(
			lambda y, a: y.eq(a.as_signed()),
			[Const(0b11, unsigned(2))], Const(0b1111, signed(4))
		)

	def test_as_signed_lhs(self):
		self.assertStatement(
			lambda y, a: y.as_signed().eq(a),
			[Const(0b01, unsigned(2))], Const(0b0001, signed(4))
		)

	def test_any(self):
		self.assertStatement(
			lambda y, a: y.eq(a.any()),
			[Const(0b00, 2)], Const(0)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.any()),
			[Const(0b01, 2)], Const(1)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.any()),
			[Const(0b10, 2)], Const(1)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.any()),
			[Const(0b11, 2)], Const(1)
		)

	def test_all(self):
		self.assertStatement(
			lambda y, a: y.eq(a.all()),
			[Const(0b00, 2)], Const(0)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.all()),
			[Const(0b01, 2)], Const(0)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.all()),
			[Const(0b10, 2)], Const(0)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.all()),
			[Const(0b11, 2)], Const(1)
		)

	def test_xor_unary(self):
		self.assertStatement(
			lambda y, a: y.eq(a.xor()),
			[Const(0b00, 2)], Const(0)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.xor()),
			[Const(0b01, 2)], Const(1)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.xor()),
			[Const(0b10, 2)], Const(1)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.xor()),
			[Const(0b11, 2)], Const(0)
		)

	def test_add(self):
		self.assertStatement(
			lambda y, a, b: y.eq(a + b),
			[Const(0,  4), Const(1, 4)], Const(1,   4)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a + b),
			[Const(-5, 4), Const(-5, 4)], Const(-10, 5)
		)

	def test_sub(self):
		self.assertStatement(
			lambda y, a, b: y.eq(a - b),
			[Const(2, 4), Const(1, 4)], Const(1,   4)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a - b),
			[Const(0, 4), Const(1, 4)], Const(-1,  4)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a - b),
			[Const(0, 4), Const(10, 4)], Const(-10, 5)
		)

	def test_mul(self):
		self.assertStatement(
			lambda y, a, b: y.eq(a * b),
			[Const(2, 4), Const(1, 4)], Const(2, 8)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a * b),
			[Const(2, 4), Const(2, 4)], Const(4, 8)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a * b),
			[Const(7, 4), Const(7, 4)], Const(49, 8)
		)

	def test_floordiv(self):
		self.assertStatement(
			lambda y, a, b: y.eq(a // b),
			[Const(2, 4), Const(1, 4)], Const(2, 8)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a // b),
			[Const(2, 4), Const(2, 4)], Const(1, 8)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a // b),
			[Const(7, 4), Const(2, 4)], Const(3, 8)
		)

	def test_floordiv_neg(self):
		self.assertStatement(
			lambda y, a, b: y.eq(a // b),
			[Const(-5, 4), Const(2, 4)], Const(-3, 8)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a // b),
			[Const(-5, 4), Const(-2, 4)], Const(2, 8)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a // b),
			[Const(5, 4), Const(2, 4)], Const(2, 8)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a // b),
			[Const(5, 4), Const(-2, 4)], Const(-3, 8)
		)

	def test_mod(self):
		self.assertStatement(
			lambda y, a, b: y.eq(a % b),
			[Const(2, 4), Const(0, 4)], Const(0, 8)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a % b),
			[Const(2, 4), Const(1, 4)], Const(0, 8)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a % b),
			[Const(2, 4), Const(2, 4)], Const(0, 8)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a % b),
			[Const(7, 4), Const(2, 4)], Const(1, 8)
		)

	def test_mod_neg(self):
		self.assertStatement(
			lambda y, a, b: y.eq(a % b),
			[Const(-5, 4), Const(3, 4)], Const(1, 8)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a % b),
			[Const(-5, 4), Const(-3, 4)], Const(-2, 8)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a % b),
			[Const(5, 4), Const(3, 4)], Const(2, 8)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a % b),
			[Const(5, 4), Const(-3, 4)], Const(-1, 8)
		)

	def test_and(self):
		self.assertStatement(
			lambda y, a, b: y.eq(a & b),
			[Const(0b1100, 4), Const(0b1010, 4)], Const(0b1000, 4)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a & b),
			[Const(0b1010, 4), Const(0b10, signed(2))], Const(0b1010, 4)
		)
		self.assertStatement(
			lambda y, a: y.eq(a),
			[Const(0b1010, 4) & Const(-2, 2).as_unsigned()], Const(0b0010, 4)
		)


	def test_or(self):
		self.assertStatement(
			lambda y, a, b: y.eq(a | b),
			[Const(0b1100, 4), Const(0b1010, 4)], Const(0b1110, 4)
		)

	def test_xor_binary(self):
		self.assertStatement(
			lambda y, a, b: y.eq(a ^ b),
			[Const(0b1100, 4), Const(0b1010, 4)], Const(0b0110, 4)
		)

	def test_shl(self):
		self.assertStatement(
			lambda y, a, b: y.eq(a << b),
			[Const(0b1001, 4), Const(0)],  Const(0b1001, 5)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a << b),
			[Const(0b1001, 4), Const(3)],  Const(0b1001000, 7)
		)

	def test_shr(self):
		self.assertStatement(
			lambda y, a, b: y.eq(a >> b),
			[Const(0b1001, 4), Const(0)],  Const(0b1001, 4)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a >> b),
			[Const(0b1001, 4), Const(2)],  Const(0b10,  4)
		)

	def test_eq(self):
		self.assertStatement(
			lambda y, a, b: y.eq(a == b),
			[Const(0, 4), Const(0, 4)], Const(1)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a == b),
			[Const(0, 4), Const(1, 4)], Const(0)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a == b),
			[Const(1, 4), Const(0, 4)], Const(0)
		)

	def test_ne(self):
		self.assertStatement(
			lambda y, a, b: y.eq(a != b),
			[Const(0, 4), Const(0, 4)], Const(0)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a != b),
			[Const(0, 4), Const(1, 4)], Const(1)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a != b),
			[Const(1, 4), Const(0, 4)], Const(1)
		)

	def test_lt(self):
		self.assertStatement(
			lambda y, a, b: y.eq(a < b),
			[Const(0, 4), Const(0, 4)], Const(0)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a < b),
			[Const(0, 4), Const(1, 4)], Const(1)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a < b),
			[Const(1, 4), Const(0, 4)], Const(0)
		)

	def test_ge(self):
		self.assertStatement(
			lambda y, a, b: y.eq(a >= b),
			[Const(0, 4), Const(0, 4)], Const(1)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a >= b),
			[Const(0, 4), Const(1, 4)], Const(0)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a >= b),
			[Const(1, 4), Const(0, 4)], Const(1)
		)

	def test_gt(self):
		self.assertStatement(
			lambda y, a, b: y.eq(a > b),
			[Const(0, 4), Const(0, 4)], Const(0)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a > b),
			[Const(0, 4), Const(1, 4)], Const(0)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a > b),
			[Const(1, 4), Const(0, 4)], Const(1)
		)

	def test_le(self):
		self.assertStatement(
			lambda y, a, b: y.eq(a <= b),
			[Const(0, 4), Const(0, 4)], Const(1)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a <= b),
			[Const(0, 4), Const(1, 4)], Const(1)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a <= b),
			[Const(1, 4), Const(0, 4)], Const(0)
		)

	def test_mux(self):
		self.assertStatement(
			lambda y, a, b, c: y.eq(Mux(c, a, b)),
			[Const(2, 4), Const(3, 4), Const(0)], Const(3, 4)
		)
		self.assertStatement(
			lambda y, a, b, c: y.eq(Mux(c, a, b)),
			[Const(2, 4), Const(3, 4), Const(1)], Const(2, 4)
		)
		self.assertStatement(
			lambda y, a: y.eq(a),
			[Mux(0, Const(0b1010, 4), Const(0b10, 2).as_signed())], Const(0b1110, 4)
		)
		self.assertStatement(
			lambda y, a: y.eq(a),
			[Mux(0, Const(0b1010, 4), Const(-2, 2).as_unsigned())], Const(0b0010, 4)
		)


	def test_mux_invert(self):
		self.assertStatement(
			lambda y, a, b, c: y.eq(Mux(~c, a, b)),
			[Const(2, 4), Const(3, 4), Const(0)], Const(2, 4)
		)
		self.assertStatement(
			lambda y, a, b, c: y.eq(Mux(~c, a, b)),
			[Const(2, 4), Const(3, 4), Const(1)], Const(3, 4)
		)

	def test_mux_wide(self):
		self.assertStatement(
			lambda y, a, b, c: y.eq(Mux(c, a, b)),
			[Const(2, 4), Const(3, 4), Const(0, 2)], Const(3, 4)
		)
		self.assertStatement(
			lambda y, a, b, c: y.eq(Mux(c, a, b)),
			[Const(2, 4), Const(3, 4), Const(1, 2)], Const(2, 4)
		)
		self.assertStatement(
			lambda y, a, b, c: y.eq(Mux(c, a, b)),
			[Const(2, 4), Const(3, 4), Const(2, 2)], Const(2, 4)
		)

	def test_abs(self):
		self.assertStatement(
			lambda y, a: y.eq(abs(a)),
			[Const(3,  unsigned(8))], Const(3, unsigned(8))
		)
		self.assertStatement(
			lambda y, a: y.eq(abs(a)),
			[Const(-3, unsigned(8))], Const(-3, unsigned(8))
		)
		self.assertStatement(
			lambda y, a: y.eq(abs(a)),
			[Const(3,  signed(8))], Const(3, signed(8))
		)
		self.assertStatement(
			lambda y, a: y.eq(abs(a)),
			[Const(-3, signed(8))], Const(3, signed(8))
		)

	def test_slice(self):
		self.assertStatement(
			lambda y, a: y.eq(a[2]),
			[Const(0b10110100, 8)], Const(0b1,  1)
		)
		self.assertStatement(
			lambda y, a: y.eq(a[2:4]),
			[Const(0b10110100, 8)], Const(0b01, 2)
		)

	def test_slice_lhs(self):
		self.assertStatement(
			lambda y, a: y[2].eq(a),
			[Const(0b0,  1)], Const(0b11111011, 8),
			reset = 0b11111111
		)
		self.assertStatement(
			lambda y, a: y[2:4].eq(a),
			[Const(0b01, 2)], Const(0b11110111, 8),
			reset = 0b11111011
		)

	def test_bit_select(self):
		self.assertStatement(
			lambda y, a, b: y.eq(a.bit_select(b, 3)),
			[Const(0b10110100, 8), Const(0)], Const(0b100, 3)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a.bit_select(b, 3)),
			[Const(0b10110100, 8), Const(2)], Const(0b101, 3)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a.bit_select(b, 3)),
			[Const(0b10110100, 8), Const(3)], Const(0b110, 3)
		)

	def test_bit_select_lhs(self):
		self.assertStatement(
			lambda y, a, b: y.bit_select(a, 3).eq(b),
			[Const(0), Const(0b100, 3)], Const(0b11111100, 8),
			reset = 0b11111111
		)
		self.assertStatement(
			lambda y, a, b: y.bit_select(a, 3).eq(b),
			[Const(2), Const(0b101, 3)], Const(0b11110111, 8),
			reset = 0b11111111
		)
		self.assertStatement(
			lambda y, a, b: y.bit_select(a, 3).eq(b),
			[Const(3), Const(0b110, 3)], Const(0b11110111, 8),
			reset = 0b11111111
		)

	def test_word_select(self):
		self.assertStatement(
			lambda y, a, b: y.eq(a.word_select(b, 3)),
			[Const(0b10110100, 8), Const(0)], Const(0b100, 3)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a.word_select(b, 3)),
			[Const(0b10110100, 8), Const(1)], Const(0b110, 3)
		)
		self.assertStatement(
			lambda y, a, b: y.eq(a.word_select(b, 3)),
			[Const(0b10110100, 8), Const(2)], Const(0b010, 3)
		)

	def test_word_select_lhs(self):
		self.assertStatement(
			lambda y, a, b: y.word_select(a, 3).eq(b),
			[Const(0), Const(0b100, 3)], Const(0b11111100, 8), reset = 0b11111111
		)
		self.assertStatement(
			lambda y, a, b: y.word_select(a, 3).eq(b),
			[Const(1), Const(0b101, 3)], Const(0b11101111, 8), reset = 0b11111111
		)
		self.assertStatement(
			lambda y, a, b: y.word_select(a, 3).eq(b),
			[Const(2), Const(0b110, 3)], Const(0b10111111, 8), reset = 0b11111111
		)

	def test_cat(self):
		self.assertStatement(
			lambda y, *xs: y.eq(Cat(*xs)),
			[Const(0b10, 2), Const(0b01, 2)], Const(0b0110, 4)
		)

	def test_cat_lhs(self):
		l = Signal(3) # noqa: E741
		m = Signal(3) # noqa: E741
		n = Signal(3) # noqa: E741
		self.assertStatement(
			lambda y, a: [Cat(l, m, n).eq(a), y.eq(Cat(n, m, l))],
			[Const(0b100101110, 9)], Const(0b110101100, 9)
		)

	def test_nested_cat_lhs(self):
		l = Signal(3) # noqa: E741
		m = Signal(3) # noqa: E741
		n = Signal(3) # noqa: E741
		self.assertStatement(
			lambda y, a: [Cat(Cat(l, Cat(m)), n).eq(a), y.eq(Cat(n, m, l))],
			[Const(0b100101110, 9)], Const(0b110101100, 9)
		)

	def test_record(self):
		rec = Record([
			('l', 1),
			('m', 2),
		])
		self.assertStatement(
			lambda y, a: [rec.eq(a), y.eq(rec)],
			[Const(0b101, 3)], Const(0b101, 3)
		)

	def test_replicate(self):
		self.assertStatement(
			lambda y, a: y.eq(a.replicate(3)),
			[Const(0b10, 2)], Const(0b101010, 6)
		)

	def test_array(self):
		array = Array([1, 4, 10])
		self.assertStatement(
			lambda y, a: y.eq(array[a]),
			[Const(0)], Const(1)
		)
		self.assertStatement(
			lambda y, a: y.eq(array[a]),
			[Const(1)], Const(4)
		)
		self.assertStatement(
			lambda y, a: y.eq(array[a]),
			[Const(2)], Const(10)
		)

	def test_array_oob(self):
		array = Array([1, 4, 10])
		self.assertStatement(
			lambda y, a: y.eq(array[a]),
			[Const(3)], Const(10)
		)
		self.assertStatement(
			lambda y, a: y.eq(array[a]),
			[Const(4)], Const(10)
		)

	def test_array_lhs(self):
		l = Signal(3, reset = 1) # noqa: E741
		m = Signal(3, reset = 4) # noqa: E741
		n = Signal(3, reset = 7) # noqa: E741

		array = Array([l, m, n])
		self.assertStatement(
			lambda y, a, b: [array[a].eq(b), y.eq(Cat(*array))],
			[Const(0), Const(0b000)], Const(0b111100000)
		)
		self.assertStatement(
			lambda y, a, b: [array[a].eq(b), y.eq(Cat(*array))],
			[Const(1), Const(0b010)], Const(0b111010001)
		)
		self.assertStatement(
			lambda y, a, b: [array[a].eq(b), y.eq(Cat(*array))],
			[Const(2), Const(0b100)], Const(0b100100001)
		)

	def test_array_lhs_oob(self):
		l = Signal(3) # noqa: E741
		m = Signal(3) # noqa: E741
		n = Signal(3) # noqa: E741
		array = Array([l, m, n])
		self.assertStatement(
			lambda y, a, b: [array[a].eq(b), y.eq(Cat(*array))],
			[Const(3), Const(0b001)], Const(0b001000000)
		)
		self.assertStatement(
			lambda y, a, b: [array[a].eq(b), y.eq(Cat(*array))],
			[Const(4), Const(0b010)], Const(0b010000000)
		)

	def test_array_index(self):
		array = Array(Array(x * y for y in range(10)) for x in range(10))
		for x in range(10):
			for y in range(10):
				self.assertStatement(
					lambda y, a, b: y.eq(array[a][b]),
					[Const(x), Const(y)], Const(x * y)
				)

	def test_array_attr(self):
		from collections import namedtuple
		pair = namedtuple('pair', ('p', 'n'))

		array = Array(pair(x, -x) for x in range(10))
		for i in range(10):
			self.assertStatement(
				lambda y, a: y.eq(array[a].p + array[a].n),
				[Const(i)], Const(0)
			)

	def test_shift_left(self):
		self.assertStatement(
			lambda y, a: y.eq(a.shift_left(1)),
			[Const(0b10100010, 8)], Const(   0b101000100, 9)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.shift_left(4)),
			[Const(0b10100010, 8)], Const(0b101000100000, 12)
		)

	def test_shift_right(self):
		self.assertStatement(
			lambda y, a: y.eq(a.shift_right(1)),
			[Const(0b10100010, 8)], Const(0b1010001, 7)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.shift_right(4)),
			[Const(0b10100010, 8)], Const(   0b1010, 4)
		)

	def test_rotate_left(self):
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_left(1)),
			[Const(0b1)], Const(0b1)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_left(1)),
			[Const(0b1001000)], Const(0b0010001)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_left(5)),
			[Const(0b1000000)], Const(0b0010000)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_left(5)),
			[Const(0b1000001)], Const(0b0110000)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_left(7)),
			[Const(0b1000000)], Const(0b1000000)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_left(7)),
			[Const(0b1000001)], Const(0b1000001)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_left(9)),
			[Const(0b1000000)], Const(0b0000010)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_left(9)),
			[Const(0b1000001)], Const(0b0000110)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_left(-1)),
			[Const(0b1)], Const(0b1)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_left(-1)),
			[Const(0b1001000)], Const(0b0100100)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_left(-5)),
			[Const(0b1000000)], Const(0b0000010)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_left(-5)),
			[Const(0b1000001)], Const(0b0000110)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_left(-7)),
			[Const(0b1000000)], Const(0b1000000)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_left(-7)),
			[Const(0b1000001)], Const(0b1000001)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_left(-9)),
			[Const(0b1000000)], Const(0b0010000)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_left(-9)),
			[Const(0b1000001)], Const(0b0110000)
		)

	def test_rotate_right(self):

		self.assertStatement(
			lambda y, a: y.eq(a.rotate_right(1)),
			[Const(0b1)], Const(0b1)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_right(1)),
			[Const(0b1001000)], Const(0b0100100)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_right(5)),
			[Const(0b1000000)], Const(0b0000010)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_right(5)),
			[Const(0b1000001)], Const(0b0000110)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_right(7)),
			[Const(0b1000000)], Const(0b1000000)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_right(7)),
			[Const(0b1000001)], Const(0b1000001)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_right(9)),
			[Const(0b1000000)], Const(0b0010000)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_right(9)),
			[Const(0b1000001)], Const(0b0110000)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_right(-1)),
			[Const(0b1)], Const(0b1)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_right(-1)),
			[Const(0b1001000)], Const(0b0010001)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_right(-5)),
			[Const(0b1000000)], Const(0b0010000)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_right(-5)),
			[Const(0b1000001)], Const(0b0110000)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_right(-7)),
			[Const(0b1000000)], Const(0b1000000)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_right(-7)),
			[Const(0b1000001)], Const(0b1000001)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_right(-9)),
			[Const(0b1000000)], Const(0b0000010)
		)
		self.assertStatement(
			lambda y, a: y.eq(a.rotate_right(-9)),
			[Const(0b1000001)], Const(0b0000110)
		)


class SimulatorIntegrationTestCase(ToriiTestSuiteCase):
	@contextmanager
	def assertSimulation(self, module, deadline = None):
		sim = Simulator(module)
		yield sim
		with sim.write_vcd('test.vcd', 'test.gtkw'):
			if deadline is None:
				sim.run()
			else:
				sim.run_until(deadline)

	def setUp_counter(self):
		self.count = Signal(3, reset = 4)
		self.sync  = ClockDomain()

		self.m = Module()
		self.m.d.sync  += self.count.eq(self.count + 1)
		self.m.domains += self.sync

	def test_counter_process(self):
		self.setUp_counter()
		with self.assertSimulation(self.m) as sim:

			def process():
				self.assertEqual((yield self.count), 4)
				yield Delay(1e-6)
				self.assertEqual((yield self.count), 4)
				yield self.sync.clk.eq(1)
				self.assertEqual((yield self.count), 4)
				yield Settle()
				self.assertEqual((yield self.count), 5)
				yield Delay(1e-6)
				self.assertEqual((yield self.count), 5)
				yield self.sync.clk.eq(0)
				self.assertEqual((yield self.count), 5)
				yield Settle()
				self.assertEqual((yield self.count), 5)
				for _ in range(3):
					yield Delay(1e-6)
					yield self.sync.clk.eq(1)
					yield Delay(1e-6)
					yield self.sync.clk.eq(0)
				self.assertEqual((yield self.count), 0)

			sim.add_process(process)

	def test_counter_clock_and_sync_process(self):
		self.setUp_counter()
		with self.assertSimulation(self.m) as sim:
			sim.add_clock(1e-6, domain = 'sync')

			def process():
				self.assertEqual((yield self.count), 4)
				self.assertEqual((yield self.sync.clk), 1)
				yield
				self.assertEqual((yield self.count), 5)
				self.assertEqual((yield self.sync.clk), 1)
				for _ in range(3):
					yield
				self.assertEqual((yield self.count), 0)

			sim.add_sync_process(process)

	def test_reset(self):
		self.setUp_counter()
		sim = Simulator(self.m)
		sim.add_clock(1e-6)
		times = 0

		def process():
			nonlocal times
			self.assertEqual((yield self.count), 4)
			yield
			self.assertEqual((yield self.count), 5)
			yield
			self.assertEqual((yield self.count), 6)
			yield
			times += 1

		sim.add_sync_process(process)
		sim.run()
		sim.reset()
		sim.run()
		self.assertEqual(times, 2)

	def setUp_alu(self):
		self.a = Signal(8)
		self.b = Signal(8)
		self.o = Signal(8)
		self.x = Signal(8)
		self.s = Signal(2)
		self.sync = ClockDomain(reset_less = True)

		self.m = Module()
		self.m.d.comb += self.x.eq(self.a ^ self.b)
		with self.m.Switch(self.s):
			with self.m.Case(0):
				self.m.d.sync += self.o.eq(self.a + self.b)
			with self.m.Case(1):
				self.m.d.sync += self.o.eq(self.a - self.b)
			with self.m.Case():
				self.m.d.sync += self.o.eq(0)
		self.m.domains += self.sync

	def test_alu(self):
		self.setUp_alu()
		with self.assertSimulation(self.m) as sim:
			sim.add_clock(1e-6)

			def process():
				yield self.a.eq(5)
				yield self.b.eq(1)
				yield
				self.assertEqual((yield self.x), 4)
				yield
				self.assertEqual((yield self.o), 6)
				yield self.s.eq(1)
				yield
				yield
				self.assertEqual((yield self.o), 4)
				yield self.s.eq(2)
				yield
				yield
				self.assertEqual((yield self.o), 0)

			sim.add_sync_process(process)

	def setUp_clock_phase(self):
		self.m = Module()
		self.phase0 = self.m.domains.phase0 = ClockDomain()
		self.phase90 = self.m.domains.phase90 = ClockDomain()
		self.phase180 = self.m.domains.phase180 = ClockDomain()
		self.phase270 = self.m.domains.phase270 = ClockDomain()
		self.check = self.m.domains.check = ClockDomain()

		self.expected = [
			[0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0],
			[0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1],
			[0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 1, 1],
			[0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0],
		]

	def test_clock_phase(self):
		self.setUp_clock_phase()
		with self.assertSimulation(self.m) as sim:
			period = 1
			sim.add_clock(period/8, phase = 0,          domain = 'check')
			sim.add_clock(period,   phase = 0*period/4, domain = 'phase0')
			sim.add_clock(period,   phase = 1*period/4, domain = 'phase90')
			sim.add_clock(period,   phase = 2*period/4, domain = 'phase180')
			sim.add_clock(period,   phase = 3*period/4, domain = 'phase270')

			def proc():
				clocks = [
					self.phase0.clk,
					self.phase90.clk,
					self.phase180.clk,
					self.phase270.clk
				]
				for i in range(16):
					yield
					for j, c in enumerate(clocks):
						self.assertEqual((yield c), self.expected[j][i])

			sim.add_sync_process(proc, domain = 'check')

	def setUp_multiclock(self):
		self.sys = ClockDomain()
		self.pix = ClockDomain()

		self.m = Module()
		self.m.domains += self.sys, self.pix

	def test_multiclock(self):
		self.setUp_multiclock()
		with self.assertSimulation(self.m) as sim:
			sim.add_clock(1e-6, domain = 'sys')
			sim.add_clock(0.3e-6, domain = 'pix')

			def sys_process():
				yield Passive()
				yield
				yield
				self.fail()

			def pix_process():
				yield
				yield
				yield

			sim.add_sync_process(sys_process, domain = 'sys')
			sim.add_sync_process(pix_process, domain = 'pix')

	def setUp_lhs_rhs(self):
		self.i = Signal(8)
		self.o = Signal(8)

		self.m = Module()
		self.m.d.comb += self.o.eq(self.i)

	def test_complex_lhs_rhs(self):
		self.setUp_lhs_rhs()
		with self.assertSimulation(self.m) as sim:
			def process():
				yield self.i.eq(0b10101010)
				yield self.i[:4].eq(-1)
				yield Settle()
				self.assertEqual((yield self.i[:4]), 0b1111)
				self.assertEqual((yield self.i), 0b10101111)
			sim.add_process(process)

	def test_run_until(self):
		m = Module()
		s = Signal()
		m.d.sync += s.eq(0)
		with self.assertSimulation(m, deadline = 100e-6) as sim:
			sim.add_clock(1e-6)

			def process():
				for _ in range(101):
					yield Delay(1e-6)
				self.fail()
			sim.add_process(process)

	def test_run_until_fail(self):
		m = Module()
		s = Signal()
		m.d.sync += s.eq(0)
		with self.assertRaises(AssertionError):
			with self.assertSimulation(m, deadline = 100e-6) as sim:
				sim.add_clock(1e-6)

				def process():
					for _ in range(99):
						yield Delay(1e-6)
					self.fail()
				sim.add_process(process)

	def test_add_process_wrong(self):
		with self.assertSimulation(Module()) as sim:
			with self.assertRaisesRegex(
				TypeError,
				r'^Cannot add a process 1 because it is not a generator function$'
			):
				sim.add_process(1)

	def test_add_process_wrong_generator(self):
		with self.assertSimulation(Module()) as sim:
			with self.assertRaisesRegex(
				TypeError,
				r'^Cannot add a process <.+?> because it is not a generator function$'
			):
				def process():
					yield Delay()
				sim.add_process(process())

	def test_add_clock_wrong_twice(self):
		m = Module()
		s = Signal()
		m.d.sync += s.eq(0)
		with self.assertSimulation(m) as sim:
			sim.add_clock(1)
			with self.assertRaisesRegex(
				ValueError,
				r'^Domain \'sync\' already has a clock driving it$'
			):
				sim.add_clock(1)

	def test_add_clock_wrong_missing(self):
		m = Module()
		with self.assertSimulation(m) as sim:
			with self.assertRaisesRegex(
				ValueError,
				r'^Domain \'sync\' is not present in simulation$'
			):
				sim.add_clock(1)

	def test_add_clock_if_exists(self):
		m = Module()
		with self.assertSimulation(m) as sim:
			sim.add_clock(1, if_exists = True)

	def test_command_wrong(self):
		survived = False
		with self.assertSimulation(Module()) as sim:
			def process():
				nonlocal survived

				with self.assertRaisesRegex(
					TypeError,
					r'Received unsupported command 1 from process .+'
				):
					yield 1
				yield Settle()
				survived = True
			sim.add_process(process)
		self.assertTrue(survived)

	def setUp_memory(self, rd_synchronous = True, rd_transparent = True, wr_granularity = None):
		self.m = Module()
		self.memory = Memory(width = 8, depth = 4, init = [0xaa, 0x55])
		self.m.submodules.rdport = self.rdport = self.memory.read_port(
			domain = 'sync' if rd_synchronous else 'comb',
			transparent = rd_transparent
		)
		self.m.submodules.wrport = self.wrport = self.memory.write_port(
			granularity = wr_granularity
		)

	def test_memory_init(self):
		self.setUp_memory()
		with self.assertSimulation(self.m) as sim:
			def process():
				yield self.rdport.addr.eq(1)
				yield
				yield
				self.assertEqual((yield self.rdport.data), 0x55)
				yield self.rdport.addr.eq(2)
				yield
				yield
				self.assertEqual((yield self.rdport.data), 0x00)
			sim.add_clock(1e-6)
			sim.add_sync_process(process)

	def test_memory_write(self):
		self.setUp_memory()
		with self.assertSimulation(self.m) as sim:
			def process():
				yield self.wrport.addr.eq(4)
				yield self.wrport.data.eq(0x33)
				yield self.wrport.en.eq(1)
				yield
				yield self.wrport.en.eq(0)
				yield self.rdport.addr.eq(4)
				yield
				self.assertEqual((yield self.rdport.data), 0x33)
			sim.add_clock(1e-6)
			sim.add_sync_process(process)

	def test_memory_write_granularity(self):
		self.setUp_memory(wr_granularity = 4)
		with self.assertSimulation(self.m) as sim:
			def process():
				yield self.wrport.data.eq(0x50)
				yield self.wrport.en.eq(0b00)
				yield
				yield self.wrport.en.eq(0)
				yield
				self.assertEqual((yield self.rdport.data), 0xaa)
				yield self.wrport.en.eq(0b10)
				yield
				yield self.wrport.en.eq(0)
				yield
				self.assertEqual((yield self.rdport.data), 0x5a)
				yield self.wrport.data.eq(0x33)
				yield self.wrport.en.eq(0b01)
				yield
				yield self.wrport.en.eq(0)
				yield
				self.assertEqual((yield self.rdport.data), 0x53)
			sim.add_clock(1e-6)
			sim.add_sync_process(process)

	def test_memory_read_before_write(self):
		self.setUp_memory(rd_transparent = False)
		with self.assertSimulation(self.m) as sim:
			def process():
				yield self.wrport.data.eq(0x33)
				yield self.wrport.en.eq(1)
				yield
				self.assertEqual((yield self.rdport.data), 0xaa)
				yield
				self.assertEqual((yield self.rdport.data), 0xaa)
				yield Settle()
				self.assertEqual((yield self.rdport.data), 0x33)
			sim.add_clock(1e-6)
			sim.add_sync_process(process)

	def test_memory_write_through(self):
		self.setUp_memory(rd_transparent = True)
		with self.assertSimulation(self.m) as sim:
			def process():
				yield self.wrport.data.eq(0x33)
				yield self.wrport.en.eq(1)
				yield
				self.assertEqual((yield self.rdport.data), 0xaa)
				yield Settle()
				self.assertEqual((yield self.rdport.data), 0x33)
				yield
				yield self.rdport.addr.eq(1)
				yield Settle()
				self.assertEqual((yield self.rdport.data), 0x33)
			sim.add_clock(1e-6)
			sim.add_sync_process(process)

	def test_memory_async_read_write(self):
		self.setUp_memory(rd_synchronous = False)
		with self.assertSimulation(self.m) as sim:
			def process():
				yield self.rdport.addr.eq(0)
				yield Settle()
				self.assertEqual((yield self.rdport.data), 0xaa)
				yield self.rdport.addr.eq(1)
				yield Settle()
				self.assertEqual((yield self.rdport.data), 0x55)
				yield self.rdport.addr.eq(0)
				yield self.wrport.addr.eq(0)
				yield self.wrport.data.eq(0x33)
				yield self.wrport.en.eq(1)
				yield Tick('sync')
				self.assertEqual((yield self.rdport.data), 0xaa)
				yield Settle()
				self.assertEqual((yield self.rdport.data), 0x33)
			sim.add_clock(1e-6)
			sim.add_process(process)

	def test_memory_read_only(self):
		self.m = Module()
		self.memory = Memory(width = 8, depth = 4, init = [0xaa, 0x55])
		self.m.submodules.rdport = self.rdport = self.memory.read_port()
		with self.assertSimulation(self.m) as sim:
			def process():
				yield
				self.assertEqual((yield self.rdport.data), 0xaa)
				yield self.rdport.addr.eq(1)
				yield
				yield
				self.assertEqual((yield self.rdport.data), 0x55)
			sim.add_clock(1e-6)
			sim.add_sync_process(process)

	def test_memory_transparency_simple(self):
		m = Module()
		init = (0x11, 0x22, 0x33, 0x44)
		m.submodules.memory = memory = Memory(width = 8, depth = 4, init = init)
		rdport = memory.read_port()
		wrport = memory.write_port(granularity = 8)
		with self.assertSimulation(m) as sim:
			def process():
				yield rdport.addr.eq(0)
				yield
				yield Settle()
				self.assertEqual((yield rdport.data), 0x11)
				yield rdport.addr.eq(1)
				yield
				yield Settle()
				self.assertEqual((yield rdport.data), 0x22)
				yield wrport.addr.eq(0)
				yield wrport.data.eq(0x44444444)
				yield wrport.en.eq(1)
				yield
				yield Settle()
				self.assertEqual((yield rdport.data), 0x22)
				yield wrport.addr.eq(1)
				yield wrport.data.eq(0x55)
				yield wrport.en.eq(1)
				yield
				yield Settle()
				self.assertEqual((yield rdport.data), 0x55)
				yield wrport.addr.eq(1)
				yield wrport.data.eq(0x66)
				yield wrport.en.eq(1)
				yield rdport.en.eq(0)
				yield
				yield Settle()
				self.assertEqual((yield rdport.data), 0x55)
				yield wrport.addr.eq(2)
				yield wrport.data.eq(0x77)
				yield wrport.en.eq(1)
				yield rdport.en.eq(1)
				yield
				yield Settle()
				self.assertEqual((yield rdport.data), 0x66)
			sim.add_clock(1e-6)
			sim.add_sync_process(process)

	def test_memory_transparency_multibit(self):
		m = Module()
		init = (0x11111111, 0x22222222, 0x33333333, 0x44444444)
		m.submodules.memory = memory = Memory(width = 32, depth = 4, init = init)
		rdport = memory.read_port()
		wrport = memory.write_port(granularity = 8)
		with self.assertSimulation(m) as sim:
			def process():
				yield rdport.addr.eq(0)
				yield
				yield Settle()
				self.assertEqual((yield rdport.data), 0x11111111)
				yield rdport.addr.eq(1)
				yield
				yield Settle()
				self.assertEqual((yield rdport.data), 0x22222222)
				yield wrport.addr.eq(0)
				yield wrport.data.eq(0x44444444)
				yield wrport.en.eq(1)
				yield
				yield Settle()
				self.assertEqual((yield rdport.data), 0x22222222)
				yield wrport.addr.eq(1)
				yield wrport.data.eq(0x55555555)
				yield wrport.en.eq(1)
				yield
				yield Settle()
				self.assertEqual((yield rdport.data), 0x22222255)
				yield wrport.addr.eq(1)
				yield wrport.data.eq(0x66666666)
				yield wrport.en.eq(2)
				yield rdport.en.eq(0)
				yield
				yield Settle()
				self.assertEqual((yield rdport.data), 0x22222255)
				yield wrport.addr.eq(1)
				yield wrport.data.eq(0x77777777)
				yield wrport.en.eq(4)
				yield rdport.en.eq(1)
				yield
				yield Settle()
				self.assertEqual((yield rdport.data), 0x22776655)
			sim.add_clock(1e-6)
			sim.add_sync_process(process)

	def test_sample_helpers(self):
		m = Module()
		s = Signal(2)

		def mk(x):
			y = Signal.like(x)
			m.d.comb += y.eq(x)
			return y

		p0, r0, f0, s0 = mk(Past(s, 0)), mk(Rose(s)),    mk(Fell(s)),    mk(Stable(s))
		p1, r1, f1, s1 = mk(Past(s)),    mk(Rose(s, 1)), mk(Fell(s, 1)), mk(Stable(s, 1))
		p2, r2, f2, s2 = mk(Past(s, 2)), mk(Rose(s, 2)), mk(Fell(s, 2)), mk(Stable(s, 2))
		p3, r3, f3, s3 = mk(Past(s, 3)), mk(Rose(s, 3)), mk(Fell(s, 3)), mk(Stable(s, 3))
		with self.assertSimulation(m) as sim:
			def process_gen():
				yield s.eq(0b10)
				yield
				yield
				yield s.eq(0b01)
				yield

			def process_check():
				yield
				yield
				yield

				self.assertEqual((yield p0), 0b01)
				self.assertEqual((yield p1), 0b10)
				self.assertEqual((yield p2), 0b10)
				self.assertEqual((yield p3), 0b00)

				self.assertEqual((yield s0), 0b0)
				self.assertEqual((yield s1), 0b1)
				self.assertEqual((yield s2), 0b0)
				self.assertEqual((yield s3), 0b1)

				self.assertEqual((yield r0), 0b01)
				self.assertEqual((yield r1), 0b00)
				self.assertEqual((yield r2), 0b10)
				self.assertEqual((yield r3), 0b00)

				self.assertEqual((yield f0), 0b10)
				self.assertEqual((yield f1), 0b00)
				self.assertEqual((yield f2), 0b00)
				self.assertEqual((yield f3), 0b00)
			sim.add_clock(1e-6)
			sim.add_sync_process(process_gen)
			sim.add_sync_process(process_check)

	def test_vcd_wrong_nonzero_time(self):
		s = Signal()
		m = Module()
		m.d.sync += s.eq(s)
		sim = Simulator(m)
		sim.add_clock(1e-6)
		sim.run_until(1e-5)
		with self.assertRaisesRegex(
			ValueError,
			r'^Cannot start writing waveforms after advancing simulation time$'
		):
			with open(os.path.devnull, 'w') as f:
				with sim.write_vcd(f):
					pass

	def test_no_negated_boolean_warning(self):
		m = Module()
		a = Signal()
		b = Signal()
		m.d.comb += a.eq(~(b == b))
		with catch_warnings(record=True) as warns:
			Simulator(m).run()
			self.assertEqual(warns, [])

	def test_large_expr_parser_overflow(self):
		m = Module()
		a = Signal()

		op = a
		for _ in range(50):
			op = (op ^ 1)

		op = op & op

		m.d.comb += a.eq(op)
		Simulator(m)

class SimulatorRegressionTestCase(ToriiTestSuiteCase):
	def test_bug_325(self):
		dut = Module()
		dut.d.comb += Signal().eq(Cat())
		Simulator(dut).run()

	def test_bug_473(self):
		sim = Simulator(Module())

		def process():
			self.assertEqual((yield -(Const(0b11, 2).as_signed())), 1)
		sim.add_process(process)
		sim.run()

	def test_bug_595(self):
		dut = Module()
		with dut.FSM(name = 'name with space'):
			with dut.State(0):
				pass
		sim = Simulator(dut)
		with self.assertRaisesRegex(
			NameError,
			r'^Signal \'bench\.top\.name with space_state\' contains a whitespace character$'
		):
			with open(os.path.devnull, 'w') as f:
				with sim.write_vcd(f):
					sim.run()

	def test_bug_588(self):
		dut = Module()
		a = Signal(32)
		b = Signal(32)
		z = Signal(32)
		dut.d.comb += z.eq(a << b)
		with self.assertRaisesRegex(
			OverflowError,
			r'^Value defined at .+?[\\/]test_sim\.py:\d+ is 4294967327 bits wide, '
			r'which is unlikely to simulate in reasonable time$'
		):
			Simulator(dut)

	def test_bug_566(self):
		dut = Module()
		dut.d.sync += Signal().eq(0)
		sim = Simulator(dut)
		with self.assertWarnsRegex(
			UserWarning,
			r'^Adding a clock process that drives a clock domain object named \'sync\', '
			r'which is distinct from an identically named domain in the simulated design$'
		):
			sim.add_clock(1e-6, domain = ClockDomain('sync'))

	def test_bug_826(self):
		sim = Simulator(Module())

		def process():
			self.assertEqual((yield Const(0b0000, 4) | ~Const(1, 1)), 0b0000)
			self.assertEqual((yield Const(0b1111, 4) & ~Const(1, 1)), 0b0000)
			self.assertEqual((yield Const(0b1111, 4) ^ ~Const(1, 1)), 0b1111)

		sim.add_process(process)
		sim.run()



# SPDX-License-Identifier: BSD-2-Clause
# torii: UnusedElaboratable=no

from torii             import *
from torii.sim         import *
from torii.lib.soc.csr import *
from torii.lib.soc     import event

from ....utils         import ToriiTestSuiteCase

def simulation_test(dut, process):
	sim = Simulator(dut)
	sim.add_clock(1e-6)
	sim.add_sync_process(process)
	with sim.write_vcd(vcd_file = open('test.vcd', 'w')):
		sim.run()


class EventMonitorTestCase(ToriiTestSuiteCase):
	def test_params(self):
		monitor = EventMonitor(data_width = 16, alignment = 4, trigger = 'rise')
		self.assertEqual(monitor.bus.data_width, 16)
		self.assertEqual(monitor.bus.memory_map.alignment, 4)
		self.assertEqual(monitor.src.trigger, event.Source.Trigger.RISE)

	def test_trigger_wrong(self):
		with self.assertRaisesRegex(
			ValueError,
			r'Invalid trigger mode \'foo\'; must be one of level, rise, fall'
		):
			EventMonitor(data_width = 8, trigger = 'foo')

	def test_add(self):
		monitor = EventMonitor(data_width = 8)
		sub = event.Source()
		monitor.add(sub)
		self.assertEqual(monitor.src.event_map.size, 1)
		self.assertEqual(monitor.src.event_map.index(sub), 0)

	def test_freeze(self):
		monitor = EventMonitor(data_width = 8)
		monitor.freeze()
		sub = event.Source()
		with self.assertRaisesRegex(
			ValueError,
			r'Event map has been frozen. Cannot add source.'
		):
			monitor.add(sub)

	def test_src_freeze(self):
		monitor = EventMonitor(data_width = 8)
		monitor.src
		sub = event.Source()
		with self.assertRaisesRegex(
			ValueError,
			r'Event map has been frozen. Cannot add source.'
		):
			monitor.add(sub)

	def test_bus_freeze(self):
		monitor = EventMonitor(data_width = 8)
		monitor.bus
		sub = event.Source()
		with self.assertRaisesRegex(
			ValueError,
			r'Event map has been frozen. Cannot add source.'
		):
			monitor.add(sub)

	def test_csr_regs(self):
		monitor = EventMonitor(data_width = 8)
		sub_0 = event.Source()
		sub_1 = event.Source()
		monitor.add(sub_0)
		monitor.add(sub_1)
		resources = list(monitor.bus.memory_map.resources())
		self.assertEqual(len(resources), 2)
		enable,  enable_name,  enable_range  = resources[0]
		pending, pending_name, pending_range = resources[1]
		self.assertEqual(
			(enable.width, enable.access, enable_range),
			(2, Element.Access.RW, (0, 1))
		)
		self.assertEqual(
			(pending.width, pending.access, pending_range),
			(2, Element.Access.RW, (1, 2))
		)

	def test_freeze_idempotent(self):
		monitor = EventMonitor(data_width = 8)
		src = monitor.src
		bus = monitor.bus
		monitor.freeze()
		self.assertIs(src, monitor.src)
		self.assertIs(bus, monitor.bus)


class EventMonitorSimulationTestCase(ToriiTestSuiteCase):
	def test_simple(self):
		dut = EventMonitor(data_width = 8)
		sub = event.Source()
		dut.add(sub)

		addr_enable  = 0x0
		addr_pending = 0x1

		def process():
			yield sub.i.eq(1)
			yield Delay()
			self.assertEqual((yield sub.trg), 1)
			self.assertEqual((yield dut.src.i), 0)

			yield dut.bus.addr.eq(addr_enable)
			yield dut.bus.r_stb.eq(1)
			yield
			yield dut.bus.r_stb.eq(0)
			yield
			self.assertEqual((yield dut.bus.r_data), 0b0)
			yield

			yield dut.bus.addr.eq(addr_enable)
			yield dut.bus.w_stb.eq(1)
			yield dut.bus.w_data.eq(0b1)
			yield
			yield dut.bus.w_stb.eq(0)
			yield
			yield Delay()

			self.assertEqual((yield dut.src.i), 1)
			yield sub.i.eq(0)
			yield Delay()
			self.assertEqual((yield sub.trg), 0)

			yield dut.bus.addr.eq(addr_pending)
			yield dut.bus.r_stb.eq(1)
			yield
			yield dut.bus.r_stb.eq(0)
			yield
			self.assertEqual((yield dut.bus.r_data), 0b1)
			yield

			yield dut.bus.addr.eq(addr_pending)
			yield dut.bus.w_stb.eq(1)
			yield dut.bus.w_data.eq(0b1)
			yield
			yield dut.bus.w_stb.eq(0)
			yield

			yield dut.bus.addr.eq(addr_pending)
			yield dut.bus.r_stb.eq(1)
			yield
			yield dut.bus.r_stb.eq(0)
			yield
			self.assertEqual((yield dut.bus.r_data), 0b0)

		simulation_test(dut, process)



# SPDX-License-Identifier: BSD-2-Clause

from .bus   import Decoder, Element, Interface, Multiplexer
from .event import EventMonitor

__all__ = (
	'Decoder',
	'Element',
	'EventMonitor',
	'Interface',
	'Multiplexer',
)



# SPDX-License-Identifier: BSD-2-Clause
# torii: UnusedElaboratable=no

from torii.hdl.ast import *
from torii.hdl.mem import *

from ..utils       import ToriiTestSuiteCase

class MemoryTestCase(ToriiTestSuiteCase):
	def test_name(self):
		m1 = Memory(width = 8, depth = 4)
		self.assertEqual(m1.name, 'm1')
		m2 = [Memory(width = 8, depth = 4)][0]
		self.assertEqual(m2.name, '$memory')
		m3 = Memory(width = 8, depth = 4, name = 'foo')
		self.assertEqual(m3.name, 'foo')

	def test_geometry(self):
		m = Memory(width = 8, depth = 4)
		self.assertEqual(m.width, 8)
		self.assertEqual(m.depth, 4)

	def test_geometry_wrong(self):
		with self.assertRaisesRegex(
			TypeError,
			r'^Memory width must be a non-negative integer, not -1$'
		):
			Memory(width = -1, depth = 4)
		with self.assertRaisesRegex(
			TypeError,
			r'^Memory depth must be a non-negative integer, not -1$'
		):
			Memory(width = 8, depth = -1)

	def test_init(self):
		m = Memory(width = 8, depth = 4, init = range(4))
		self.assertEqual(m.init, [0, 1, 2, 3])

	def test_init_wrong_count(self):
		with self.assertRaisesRegex(
			ValueError,
			r'^Memory initialization value count exceed memory depth \(8 > 4\)$'
		):
			Memory(width = 8, depth = 4, init = range(8))

	def test_init_wrong_type(self):
		with self.assertRaisesRegex(
			TypeError, (
				r'^Memory initialization value at address 1: '
				r'\'str\' object cannot be interpreted as an integer$'
			)
		):
			Memory(width = 8, depth = 4, init = [1, '0'])

	def test_attrs(self):
		m1 = Memory(width = 8, depth = 4)
		self.assertEqual(m1.attrs, {})
		m2 = Memory(width = 8, depth = 4, attrs = {'ram_block': True})
		self.assertEqual(m2.attrs, {'ram_block': True})

	def test_read_port_transparent(self):
		mem    = Memory(width = 8, depth = 4)
		rdport = mem.read_port()
		self.assertEqual(rdport.memory, mem)
		self.assertEqual(rdport.domain, 'sync')
		self.assertEqual(rdport.transparent, True)
		self.assertEqual(len(rdport.addr), 2)
		self.assertEqual(len(rdport.data), 8)
		self.assertEqual(len(rdport.en), 1)
		self.assertIsInstance(rdport.en, Signal)
		self.assertEqual(rdport.en.reset, 1)

	def test_read_port_non_transparent(self):
		mem    = Memory(width = 8, depth = 4)
		rdport = mem.read_port(transparent = False)
		self.assertEqual(rdport.memory, mem)
		self.assertEqual(rdport.domain, 'sync')
		self.assertEqual(rdport.transparent, False)
		self.assertEqual(len(rdport.en), 1)
		self.assertIsInstance(rdport.en, Signal)
		self.assertEqual(rdport.en.reset, 1)

	def test_read_port_asynchronous(self):
		mem    = Memory(width = 8, depth = 4)
		rdport = mem.read_port(domain = 'comb')
		self.assertEqual(rdport.memory, mem)
		self.assertEqual(rdport.domain, 'comb')
		self.assertEqual(rdport.transparent, True)
		self.assertEqual(len(rdport.en), 1)
		self.assertIsInstance(rdport.en, Const)
		self.assertEqual(rdport.en.value, 1)

	def test_read_port_wrong(self):
		mem = Memory(width = 8, depth = 4)
		with self.assertRaisesRegex(
			ValueError,
			r'^Read port cannot be simultaneously asynchronous and non-transparent$'
		):
			mem.read_port(domain = 'comb', transparent = False)

	def test_write_port(self):
		mem    = Memory(width = 8, depth = 4)
		wrport = mem.write_port()
		self.assertEqual(wrport.memory, mem)
		self.assertEqual(wrport.domain, 'sync')
		self.assertEqual(wrport.granularity, 8)
		self.assertEqual(len(wrport.addr), 2)
		self.assertEqual(len(wrport.data), 8)
		self.assertEqual(len(wrport.en), 1)

	def test_write_port_granularity(self):
		mem    = Memory(width = 8, depth = 4)
		wrport = mem.write_port(granularity = 2)
		self.assertEqual(wrport.memory, mem)
		self.assertEqual(wrport.domain, 'sync')
		self.assertEqual(wrport.granularity, 2)
		self.assertEqual(len(wrport.addr), 2)
		self.assertEqual(len(wrport.data), 8)
		self.assertEqual(len(wrport.en), 4)

	def test_write_port_granularity_wrong(self):
		mem = Memory(width = 8, depth = 4)
		with self.assertRaisesRegex(
			TypeError,
			r'^Write port granularity must be a non-negative integer, not -1$'
		):
			mem.write_port(granularity = -1)
		with self.assertRaisesRegex(
			ValueError,
			r'^Write port granularity must not be greater than memory width \(10 > 8\)$'
		):
			mem.write_port(granularity = 10)
		with self.assertRaisesRegex(
			ValueError,
			r'^Write port granularity must divide memory width evenly$'
		):
			mem.write_port(granularity = 3)


class DummyPortTestCase(ToriiTestSuiteCase):
	def test_name(self):
		p1 = DummyPort(data_width = 8, addr_width = 2)
		self.assertEqual(p1.addr.name, 'p1_addr')
		p2 = [DummyPort(data_width = 8, addr_width = 2)][0]
		self.assertEqual(p2.addr.name, 'dummy_addr')
		p3 = DummyPort(data_width = 8, addr_width = 2, name = 'foo')
		self.assertEqual(p3.addr.name, 'foo_addr')

	def test_sizes(self):
		p1 = DummyPort(data_width = 8, addr_width = 2)
		self.assertEqual(p1.addr.width, 2)
		self.assertEqual(p1.data.width, 8)
		self.assertEqual(p1.en.width, 1)
		p2 = DummyPort(data_width = 8, addr_width = 2, granularity = 2)
		self.assertEqual(p2.en.width, 4)



# SPDX-License-Identifier: BSD-2-Clause

from ..hdl.rec import DIR_FANIN, DIR_FANOUT, Record

__all__ = (
	'MockPlatform',
	'MockRecord',
)


class MockRecord(Record):
	'''
	Mock Torii Record

	This class is a modified version of :py:class:`Record`
	that dynamically allocates a record for every field
	requested by name, and caches it.

	This allows :py:class:`MockPlatform` to return an object
	of :py:class:`MockRecord` from ``request`` and have it
	always be "correct".

	This is done by shimming the ``__getitem__`` call and when
	the field lookup fails to allocate a new record with the
	requested name and return that.

	..warning:: This only returns a record with 1-wide ``o`` and ``i`` fields

	'''

	def _insert_field(self, item: str) -> None:
		'''
		Construct mocked Record

		When the call to ``__getitem__`` runs into a field that
		we have not mocked before, and the field is a str, then
		we insert a mocked record with an ``i`` and ``o`` field
		into the parent records fields.

		..warning:: The field widths are currently set at 1

		Parameters
		----------
		item : str
			The requested field to mock

		'''

		self.fields[item] = Record((
			('o', 1, DIR_FANOUT),
			('i', 1, DIR_FANIN )
		), name = f'{item}')

	def __init__(self, *args, **kwargs) -> None:
		super().__init__((), *args, **kwargs)
		self.fields = {}

	def __getitem__(self, item: str) -> Record:
		'''
		Return record field

		If the field doesn't exist, we create it with a call
		to ``_insert_field`` then return the generated field.

		Parameters
		----------
		item : Any
			The name of the requested field

		'''

		if isinstance(item, str):
			try:
				return self.fields[item]
			except KeyError:
				self._insert_field(item)
			return self.fields[item]
		else:
			return super().__getitem__(item)

class MockPlatform:
	'''
	Mock Torii Platform

	This is a mock platform that returns a :py:class:`MockRecord`
	whenever any call to ``request`` is made, regardless of the
	resource being requested.

	'''

	def request(self, name: str, number: int = 0) -> MockRecord:
		'''
		Request a resource from the platform

		This call returns a :py:class:`MockRecord` regardless
		of any of the parameters passed in.

		Parameters
		----------
		name : str
			The name of the resource to request
		number : int
			The index of the resource to request

		'''

		return MockRecord()



# SPDX-License-Identifier: BSD-2-Clause

import os
import ctypes
import tempfile

from torii.tools.cxx import build_cxx
from torii.test      import ToriiTestCase

class ToolchainCxxTestCase(ToriiTestCase):
	def setUp(self):
		self.include_dir = None
		self.build_dir = None

	def tearDown(self):
		if self.include_dir:
			self.include_dir.cleanup()
		if self.build_dir:
			self.build_dir.cleanup()

	def test_filename(self):
		self.build_dir, filename = build_cxx(
			cxx_sources = {'test.cc': ''},
			output_name = 'answer',
			include_dirs = [],
			macros = [],
		)
		self.assertTrue(filename.startswith('answer'))

	def test_simple(self):
		self.build_dir, filename = build_cxx(
			cxx_sources = {
				'test.cc': '''
					extern "C" int answer() { return 42; }
				'''
			},
			output_name = 'answer',
			include_dirs = [],
			macros = [],
		)
		library = ctypes.cdll.LoadLibrary(os.path.join(self.build_dir.name, filename))
		self.assertEqual(library.answer(), 42)

	def test_macro(self):
		self.build_dir, filename = build_cxx(
			cxx_sources = {
				'test.cc': '''
					extern "C" int answer() { return ANSWER; }
				'''
			},
			output_name = 'answer',
			include_dirs = [],
			macros = ['ANSWER = 42'],
		)
		library = ctypes.cdll.LoadLibrary(os.path.join(self.build_dir.name, filename))
		self.assertEqual(library.answer(), 42)

	def test_include(self):
		self.include_dir = tempfile.TemporaryDirectory(prefix = 'torii_hxx_')
		with open(os.path.join(self.include_dir.name, 'answer.h'), 'w') as f:
			f.write('#define ANSWER 42')

		self.build_dir, filename = build_cxx(
			cxx_sources = {
				'test.cc': '''
					#include <answer.h>
					extern "C" int answer() { return ANSWER; }
				'''
			},
			output_name = 'answer',
			include_dirs = [self.include_dir.name],
			macros = [],
		)
		library = ctypes.cdll.LoadLibrary(os.path.join(self.build_dir.name, filename))
		self.assertEqual(library.answer(), 42)



# SPDX-License-Identifier: BSD-2-Clause

from .hdl.ast import (
	AnyConst, AnySeq, Assert, Assume, Cover, Fell, Initial,
	Past, Rose, Stable
)

__all__ = (
	'AnyConst',
	'AnySeq',
	'Assert',
	'Assume',
	'Cover',
	'Fell',
	'Initial',
	'Past',
	'Rose',
	'Stable',
)



# SPDX-License-Identifier: BSD-2-Clause
# torii: UnusedElaboratable=no

from .... import Elaboratable, Module
from ..   import event
from .    import Element, Multiplexer
from .bus import Interface

__all__ = (
	'EventMonitor',
)


class EventMonitor(Elaboratable):
	'''
	Event monitor.

	A monitor for subordinate event sources, with a CSR bus interface.

	CSR registers
	-------------
	enable : ``self.src.event_map.size``, read/write
		Enabled events. See :meth:`..event.EventMap.sources` for layout.
	pending : ``self.src.event_map.size``, read/clear
		Pending events. See :meth:`..event.EventMap.sources` for layout.

	Parameters
	----------
	data_width : int
		CSR bus data width. See :class:`..csr.Interface`.
	alignment : int
		CSR address alignment. See :class:`..memory.MemoryMap`.
	trigger : :class:`..event.Source.Trigger`
		Trigger mode. See :class:`..event.Source`.

	'''

	def __init__(
		self, *, data_width: int, alignment: int = 0, trigger: event.Source.Trigger = 'level'
	) -> None:
		choices = ('level', 'rise', 'fall')
		if not isinstance(trigger, event.Source.Trigger) and trigger not in choices:
			raise ValueError(f'Invalid trigger mode {trigger!r}; must be one of {", ".join(choices)}')

		self._trigger = event.Source.Trigger(trigger)
		self._map     = event.EventMap()
		self._monitor = None
		self._enable  = None
		self._pending = None
		self._mux     = Multiplexer(addr_width = 1, data_width = data_width, alignment = alignment)
		self._frozen  = False

	def freeze(self) -> None:
		'''
		Freeze the event monitor.

		Once the event monitor is frozen, subordinate sources cannot be added anymore.

		'''

		if self._frozen:
			return
		self._monitor = event.Monitor(self._map, trigger = self._trigger)
		self._enable  = Element(self._map.size, 'rw')
		self._pending = Element(self._map.size, 'rw')
		self._mux.add(self._enable,  extend = True)
		self._mux.add(self._pending, extend = True)
		self._frozen  = True

	@property
	def src(self) -> event.Source:
		'''
		Event source.

		Return value
		------------
		An :class:`..event.Source`. Its input line is asserted by the monitor when a subordinate
		event is enabled and pending.

		'''

		self.freeze()
		return self._monitor.src

	@property
	def bus(self) -> Interface:
		'''
		CSR bus interface.

		Return value
		------------
		A :class:`..csr.Interface` providing access to registers.

		'''

		self.freeze()
		return self._mux.bus

	def add(self, src: event.Source) -> None:
		'''
		Add a subordinate event source.

		See :meth:`..event.EventMap.add` for details.

		'''

		self._map.add(src)

	def elaborate(self, platform) -> Module:
		self.freeze()

		m = Module()
		m.submodules.monitor = self._monitor
		m.submodules.mux     = self._mux

		with m.If(self._enable.w_stb):
			m.d.sync += self._monitor.enable.eq(self._enable.w_data)
		m.d.comb += self._enable.r_data.eq(self._monitor.enable)

		with m.If(self._pending.w_stb):
			m.d.comb += self._monitor.clear.eq(self._pending.w_data)
		m.d.comb += self._pending.r_data.eq(self._monitor.pending)

		return m



# SPDX-License-Identifier: BSD-2-Clause

import os
import re
import textwrap
from abc             import ABCMeta, abstractmethod
from collections     import OrderedDict
from collections.abc import Iterable
from typing          import IO, Generator, Literal, Optional, Type, TypeVar, Union

import jinja2

from ..              import __version__
from ..back          import rtlil, verilog
from ..hdl.ast       import ClockSignal, Const, Signal
from ..hdl.cd        import ClockDomain
from ..hdl.dsl       import Module
from ..hdl.ir        import Elaboratable, Fragment, Instance
from ..hdl.rec       import Record
from ..hdl.xfrm      import DomainLowerer, SampleLowerer
from ..lib.cdc       import ResetSynchronizer
from ..lib.io        import Pin
from ..tools         import has_tool, require_tool
from ..util.string   import ascii_escape, tcl_escape, tcl_quote, tool_env_var
from .dsl            import Attrs, Clock
from .res            import ResourceManager
from .run            import BuildPlan, BuildProducts

__all__ = (
	'Platform',
	'TemplatedPlatform',
)


class Platform(ResourceManager, metaclass = ABCMeta):
	@property
	@abstractmethod
	def resources(self):
		raise NotImplementedError('Platform must implement this property')

	@property
	def connectors(self):
		return [] # Connectors are very rarely used, so make declaring them optional

	default_clk    = None
	default_rst    = None

	@property
	@abstractmethod
	def required_tools(self):
		raise NotImplementedError('Platform must implement this property')


	def __init__(self) -> None:
		super().__init__(self.resources, self.connectors)

		self.extra_files = OrderedDict()

		self._prepared   = False

	@property
	def default_clk_constraint(self) -> Clock:
		if self.default_clk is None:
			raise AttributeError(f'Platform \'{type(self).__name__}\' does not define a default clock')

		return self.lookup(self.default_clk).clock

	@property
	def default_clk_frequency(self) -> float:
		constraint = self.default_clk_constraint
		if constraint is None:
			raise AttributeError(f'Platform \'{type(self).__name__}\' does not constrain its default clock')

		return constraint.frequency

	def add_file(self, filename: str, content: Union[str, bytes, IO]) -> None:
		if not isinstance(filename, str):
			raise TypeError(f'File name must be a string, not {filename!r}')
		if hasattr(content, 'read'):
			content = content.read()
		elif not isinstance(content, (str, bytes)):
			raise TypeError(f'File contents must be str, bytes, or a file-like object, not {content!r}')

		if filename in self.extra_files:
			if self.extra_files[filename] != content:
				raise ValueError(f'File {filename!r} already exists')
		else:
			self.extra_files[filename] = content

	def iter_files(self, *suffixes: str) -> Iterable[str]:
		for filename in self.extra_files:
			if filename.endswith(suffixes):
				yield filename

	@property
	def _toolchain_env_var(self) -> str:
		return f'TORII_ENV_{tool_env_var(self.toolchain)}'

	def build(
		self, elaboratable: Union[Fragment, Elaboratable], name: str = 'top', build_dir: str = 'build',
		do_build: bool = True, program_opts: Optional[dict[str, str]] = None, do_program: bool = False, **kwargs
	) -> Union[BuildPlan, BuildProducts, None]:
		# The following code performs a best-effort check for presence of required tools upfront,
		# before performing any build actions, to provide a better diagnostic. It does not handle
		# several corner cases:
		#  1. `require_tool` does not source toolchain environment scripts, so if such a script
		#     is used, the check is skipped, and `execute_local()` may fail;
		#  2. if the design is not built (do_build = False), most of the tools are not required and
		#     in fact might not be available if the design will be built manually with a different
		#     environment script specified, or on a different machine; however, Yosys is required
		#     by virtually every platform anyway, to provide debug Verilog output, and `prepare()`
		#     may fail.
		# This is OK because even if `require_tool` succeeds, the toolchain might be broken anyway.
		# The check only serves to catch common errors earlier.
		if do_build and (self._toolchain_env_var not in os.environ):
			for tool in self.required_tools:
				require_tool(tool)

		plan = self.prepare(elaboratable, name, **kwargs)
		if not do_build:
			return plan

		products = plan.execute_local(build_dir)
		if not do_program:
			return products

		self.toolchain_program(products, name, **(program_opts or {}))

	def has_required_tools(self) -> bool:
		if self._toolchain_env_var in os.environ:
			return True
		return all(has_tool(name) for name in self.required_tools)

	def create_missing_domain(self, name: str) -> Module:
		# Simple instantiation of a clock domain driven directly by the board clock and reset.
		# This implementation uses a single ResetSynchronizer to ensure that:
		#   * an external reset is definitely synchronized to the system clock;
		#   * release of power-on reset, which is inherently asynchronous, is synchronized to
		#     the system clock.
		# Many device families provide advanced primitives for tackling reset. If these exist,
		# they should be used instead.
		if name == 'sync' and self.default_clk is not None:
			clk_i = self.request(self.default_clk).i
			if self.default_rst is not None:
				rst_i = self.request(self.default_rst).i
			else:
				rst_i = Const(0)

			m = Module()
			m.domains += ClockDomain('sync')
			m.d.comb += ClockSignal('sync').eq(clk_i)
			m.submodules.reset_sync = ResetSynchronizer(rst_i, domain = 'sync')
			return m

	def prepare(
		self, elaboratable: Union[Fragment, Elaboratable], name: str = 'top', **kwargs
	) -> BuildPlan:
		assert not self._prepared
		self._prepared = True

		fragment = Fragment.get(elaboratable, self)
		fragment = SampleLowerer()(fragment)
		fragment._propagate_domains(self.create_missing_domain, platform = self)
		fragment = DomainLowerer()(fragment)

		def add_pin_fragment(pin, pin_fragment):
			pin_fragment = Fragment.get(pin_fragment, self)
			if not isinstance(pin_fragment, Instance):
				pin_fragment.flatten = True
			fragment.add_subfragment(pin_fragment, name = f'pin_{pin.name}')

		for pin, port, attrs, invert in self.iter_single_ended_pins():
			if pin.dir == 'i':
				add_pin_fragment(pin, self.get_input(pin, port, attrs, invert))
			if pin.dir == 'o':
				add_pin_fragment(pin, self.get_output(pin, port, attrs, invert))
			if pin.dir == 'oe':
				add_pin_fragment(pin, self.get_tristate(pin, port, attrs, invert))
			if pin.dir == 'io':
				add_pin_fragment(pin, self.get_input_output(pin, port, attrs, invert))

		for pin, port, attrs, invert in self.iter_differential_pins():
			if pin.dir == 'i':
				add_pin_fragment(pin, self.get_diff_input(pin, port, attrs, invert))
			if pin.dir == 'o':
				add_pin_fragment(pin, self.get_diff_output(pin, port, attrs, invert))
			if pin.dir == 'oe':
				add_pin_fragment(pin, self.get_diff_tristate(pin, port, attrs, invert))
			if pin.dir == 'io':
				add_pin_fragment(pin, self.get_diff_input_output(pin, port, attrs, invert))

		fragment._propagate_ports(ports = self.iter_ports(), all_undef_as_ports = False)
		return self.toolchain_prepare(fragment, name, **kwargs)

	@abstractmethod
	def toolchain_prepare(self, fragment: Fragment, name: str, **kwargs) -> BuildPlan:
		'''
		Convert the ``fragment`` and constraints recorded in this :class:`Platform` into
		a :class:`BuildPlan`.
		'''
		raise NotImplementedError # :nocov:

	def toolchain_program(self, products: BuildProducts, name: str, **kwargs) -> None:
		'''
		Extract bitstream for fragment ``name`` from ``products`` and download it to a target.
		'''
		raise NotImplementedError(f'Platform \'{type(self).__name__}\' does not support programming')

	def _check_feature(
		self, feature: str, pin: Pin, attrs: Attrs, valid_xdrs: tuple[int], valid_attrs: Optional[str]
	) -> None:
		if len(valid_xdrs) == 0:
			raise NotImplementedError(f'Platform \'{type(self).__name__}\' does not support {feature}')

		elif pin.xdr not in valid_xdrs:
			raise NotImplementedError(f'Platform \'{type(self).__name__}\' does not support {feature} for XDR {pin.xdr}')

		if not valid_attrs and attrs:
			raise NotImplementedError(f'Platform \'{type(self).__name__}\' does not support attributes for {feature}')

	@staticmethod
	def _invert_if(invert: bool, value):
		if invert:
			return ~value
		else:
			return value

	def get_input(
		self, pin: Pin, port: Record, attrs: Attrs, invert: bool
	) -> Module:
		self._check_feature('single-ended input', pin, attrs, valid_xdrs = (0,), valid_attrs = None)

		m = Module()
		m.d.comb += pin.i.eq(self._invert_if(invert, port))
		return m

	def get_output(
		self, pin: Pin, port: Record, attrs: Attrs, invert: bool
	) -> Module:
		self._check_feature('single-ended output', pin, attrs, valid_xdrs = (0,), valid_attrs = None)

		m = Module()
		m.d.comb += port.eq(self._invert_if(invert, pin.o))
		return m

	def get_tristate(
		self, pin: Pin, port: Record, attrs: Attrs, invert: bool
	) -> Module:
		self._check_feature('single-ended tristate', pin, attrs, valid_xdrs = (0,), valid_attrs = None)

		m = Module()
		m.submodules += Instance('$tribuf',
			p_WIDTH = pin.width,
			i_EN    = pin.oe,
			i_A     = self._invert_if(invert, pin.o),
			o_Y     = port,
		)
		return m

	def get_input_output(
		self, pin: Pin, port: Record, attrs: Attrs, invert: bool
	) -> Module:
		self._check_feature('single-ended input/output', pin, attrs, valid_xdrs = (0,), valid_attrs = None)

		m = Module()
		m.submodules += Instance('$tribuf',
			p_WIDTH = pin.width,
			i_EN    = pin.oe,
			i_A     = self._invert_if(invert, pin.o),
			o_Y     = port,
		)
		m.d.comb += pin.i.eq(self._invert_if(invert, port))
		return m

	def get_diff_input(
		self, pin: Pin, port: Record, attrs: Attrs, invert: bool
	) -> None:
		self._check_feature('differential input', pin, attrs, valid_xdrs = (), valid_attrs = None)

	def get_diff_output(
		self, pin: Pin, port: Record, attrs: Attrs, invert: bool
	) -> None:
		self._check_feature('differential output', pin, attrs, valid_xdrs = (), valid_attrs = None)

	def get_diff_tristate(
		self, pin: Pin, port: Record, attrs: Attrs, invert: bool
	) -> None:
		self._check_feature('differential tristate', pin, attrs, valid_xdrs = (), valid_attrs = None)

	def get_diff_input_output(
		self, pin: Pin, port: Record, attrs: Attrs, invert: bool
	) -> None:
		self._check_feature('differential input/output', pin, attrs, valid_xdrs = (), valid_attrs = None)


class TemplatedPlatform(Platform):
	@property
	@abstractmethod
	def toolchain(self) -> Optional[str]:
		raise NotImplementedError('Platform must implement this property')

	@property
	@abstractmethod
	def file_templates(self) -> dict[str, str]:
		raise NotImplementedError('Platform must implement this property')

	@property
	@abstractmethod
	def command_templates(self) -> list[str]:
		raise NotImplementedError('Platform must implement this property')

	build_script_templates = {
		'build_{{name}}.sh': '''
			# {{autogenerated}}
			set -e{{verbose("x")}}
			[ -n "${{platform._toolchain_env_var}}" ] && . "${{platform._toolchain_env_var}}"
			{{emit_commands("sh")}}
		''',
		'build_{{name}}.bat': '''
			@rem {{autogenerated}}
			{{quiet("@echo off")}}
			if defined {{platform._toolchain_env_var}} call %{{platform._toolchain_env_var}}%
			{{emit_commands("bat")}}
		''',
	}

	def iter_clock_constraints(self) -> Generator[
		tuple[str, Signal, float], None, None
	]:
		for net_signal, port_signal, frequency in super().iter_clock_constraints():
			# Skip any clock constraints placed on signals that are never used in the design.
			# Otherwise, it will cause a crash in the vendor platform if it supports clock
			# constraints on non-port nets.
			if net_signal not in self._name_map:
				continue
			yield (net_signal, port_signal, frequency)

	def toolchain_prepare(self, fragment: Fragment, name: str, *, emit_src = True, **kwargs) -> BuildPlan:
		# Restrict the name of the design to a strict alphanumeric character set. Platforms will
		# interpolate the name of the design in many different contexts: filesystem paths, Python
		# scripts, Tcl scripts, ad-hoc constraint files, and so on. It is not practical to add
		# escaping code that handles every one of their edge cases, so make sure we never hit them
		# in the first place.
		invalid_char = re.match(r'[^A-Za-z0-9_]', name)
		if invalid_char:
			raise ValueError(
				f'Design name {name!r} contains invalid character {invalid_char.group(0)!r}; only alphanumeric '
				'characters are valid in design names'
			)

		# This notice serves a dual purpose: to explain that the file is autogenerated,
		# and to incorporate the Torii version into generated code.
		autogenerated = f'Automatically generated by Torii {__version__}. Do not edit.'

		rtlil_text, self._name_map = rtlil.convert_fragment(fragment, name = name, emit_src = emit_src)

		# Retrieve an override specified in either the environment or as a kwarg.
		# expected_type parameter is used to assert the type of kwargs, passing `None` will disable
		# type checking.
		_ETYPE = TypeVar('_ETYPE')

		def _extract_override(
			var: str, *, expected_type: Type[_ETYPE]
		) -> Union[jinja2.Undefined, _ETYPE]:
			var_env = f'TORII_{tool_env_var(var)}'
			if var_env in os.environ:
				# On Windows, there is no way to define an "empty but set" variable; it is tempting
				# to use a quoted empty string, but it doesn't do what one would expect. Recognize
				# this as a useful pattern anyway, and treat `set VAR=""` on Windows the same way
				# `export VAR=` is treated on Linux.
				if var_env in os.environ:
					var_env_value = os.environ[var_env]

				return re.sub(r'^\"\"$', '', var_env_value)
			elif var in kwargs:
				kwarg = kwargs[var]
				if issubclass(expected_type, str) and not isinstance(kwarg, str) and isinstance(kwarg, Iterable):
					kwarg = ' '.join(kwarg)
				if not isinstance(kwarg, expected_type) and expected_type is not None:
					raise TypeError(f'Override \'{var}\' must be a {expected_type.__name__}, not {kwarg!r}')
				return kwarg
			else:
				return jinja2.Undefined(name = var)

		def get_override(var: str) -> str:
			value = _extract_override(var, expected_type = str)
			return value

		def get_override_flag(var: str) -> bool:
			value = _extract_override(var, expected_type = bool)
			if isinstance(value, str):
				value = value.lower()
				if value in ('0', 'no', 'n', 'false', ''):
					return False
				if value in ('1', 'yes', 'y', 'true'):
					return True
				else:
					raise ValueError(
						f'Override \'{var}\' must be one of ("0", "n", "no", "false", "")'
						f' or ("1", "y", "yes", "true"), not {value!r}'
					)
			return value

		def emit_rtlil() -> str:
			return rtlil_text

		def emit_verilog(opts: tuple[str] = ()) -> str:
			return verilog._convert_rtlil_text(
				rtlil_text, strip_internal_attrs = True, write_verilog_opts = opts
			)

		def emit_debug_verilog(opts: str = ()) -> str:
			if not get_override_flag('debug_verilog'):
				return '/* Debug Verilog generation was disabled. */'
			else:
				return verilog._convert_rtlil_text(
					rtlil_text, strip_internal_attrs = False, write_verilog_opts = opts
				)

		def emit_commands(syntax: Literal['sh', 'bat']) -> str:
			commands = []

			for name in self.required_tools:
				env_var = tool_env_var(name)
				if syntax == 'sh':
					template = ': ${{{env_var}:={name}}}'
				elif syntax == 'bat':
					template = \
						'if [%{env_var}%] equ [\"\"] set {env_var}=\n' \
						'if [%{env_var}%] equ [] set {env_var}={name}'
				else:
					assert False
				commands.append(template.format(env_var = env_var, name = name))

			for index, command_tpl in enumerate(self.command_templates):
				command = render(
					command_tpl, origin = f'<command#{index + 1}>', syntax = syntax
				)
				command = re.sub(r'\s+', ' ', command)
				if syntax == 'sh':
					commands.append(command)
				elif syntax == 'bat':
					commands.append(command + ' || exit /b')
				else:
					assert False

			return '\n'.join(commands)

		@jinja2.pass_context
		def invoke_tool(context, name: str) -> str:
			env_var = tool_env_var(name)
			if context.parent['syntax'] == 'sh':
				return f'"${env_var}"'
			elif context.parent['syntax'] == 'bat':
				return f'%{env_var}%'
			else:
				assert False

		def options(opts: Union[str, Iterable[str]]) -> str:
			if isinstance(opts, str):
				return opts
			else:
				return ' '.join(opts)

		def hierarchy(signal, separator):
			return separator.join(self._name_map[signal][1:])

		def verbose(arg: str) -> Union[jinja2.Undefined, str]:
			if get_override_flag('verbose'):
				return arg
			else:
				return jinja2.Undefined(name = 'quiet')

		def quiet(arg: str) -> Union[jinja2.Undefined, str]:
			if get_override_flag('verbose'):
				return jinja2.Undefined(name = 'quiet')
			else:
				return arg

		def render(
			source: str, origin: str, syntax: Optional[Literal['sh', 'bat']] = None
		) -> str:
			try:
				source   = textwrap.dedent(source).strip()
				compiled = jinja2.Template(
					source,
					trim_blocks = True,
					lstrip_blocks = True,
					undefined = jinja2.StrictUndefined
				)
				compiled.environment.filters['options']      = options
				compiled.environment.filters['hierarchy']    = hierarchy
				compiled.environment.filters['ascii_escape'] = ascii_escape
				compiled.environment.filters['tcl_escape']   = tcl_escape
				compiled.environment.filters['tcl_quote']    = tcl_quote
			except jinja2.TemplateSyntaxError as e:
				e.args = (f'{e.message} (at {origin}:{e.lineno})',)
				raise
			return compiled.render({
				'name'              : name,
				'platform'          : self,
				'emit_rtlil'        : emit_rtlil,
				'emit_verilog'      : emit_verilog,
				'emit_debug_verilog': emit_debug_verilog,
				'emit_commands'     : emit_commands,
				'syntax'            : syntax,
				'invoke_tool'       : invoke_tool,
				'get_override'      : get_override,
				'get_override_flag' : get_override_flag,
				'verbose'           : verbose,
				'quiet'             : quiet,
				'autogenerated'     : autogenerated,
			})

		plan = BuildPlan(script = f'build_{name}')
		for filename_tpl, content_tpl in self.file_templates.items():
			plan.add_file(
				render(filename_tpl, origin = filename_tpl),
				render(content_tpl, origin = content_tpl)
			)
		for filename, content in self.extra_files.items():
			plan.add_file(filename, content)
		return plan



# SPDX-License-Identifier: BSD-2-Clause

from torii         import Signal, Module
from torii.lib.cdc import FFSynchronizer
from torii.back    import verilog


i, o = Signal(name = 'i'), Signal(name = 'o')
m = Module()
m.submodules += FFSynchronizer(i, o)

if __name__ == '__main__':
	print(verilog.convert(m, ports = [i, o]))



# SPDX-License-Identifier: BSD-2-Clause

from torii      import Elaboratable, Module, Signal, Memory
from torii.back import verilog


class RegisterFile(Elaboratable):
	def __init__(self):
		self.adr   = Signal(4)
		self.dat_r = Signal(8)
		self.dat_w = Signal(8)
		self.we    = Signal()
		self.mem   = Memory(width = 8, depth = 16, init = [0xaa, 0x55])

	def elaborate(self, platform) -> Module:
		m = Module()
		m.submodules.rdport = rdport = self.mem.read_port()
		m.submodules.wrport = wrport = self.mem.write_port()
		m.d.comb += [
			rdport.addr.eq(self.adr),
			self.dat_r.eq(rdport.data),
			wrport.addr.eq(self.adr),
			wrport.data.eq(self.dat_w),
			wrport.en.eq(self.we),
		]
		return m


if __name__ == '__main__':
	rf = RegisterFile()

	print(verilog.convert(rf, ports = [rf.adr, rf.dat_r, rf.dat_w, rf.we]))



# SPDX-License-Identifier: BSD-2-Clause

from enum          import Enum

from torii.hdl.ast import *
from torii.hdl.rec import *

from ..utils       import ToriiTestSuiteCase


class UnsignedEnum(Enum):
	FOO = 1
	BAR = 2
	BAZ = 3


class LayoutTestCase(ToriiTestSuiteCase):
	def assertFieldEqual(self, field, expected):
		(shape, dir) = field
		shape = Shape.cast(shape)
		self.assertEqual((shape, dir), expected)

	def test_fields(self):
		layout = Layout.cast([
			('cyc',  1),
			('data', signed(32)),
			('stb',  1, DIR_FANOUT),
			('ack',  1, DIR_FANIN),
			('info', [
				('a', 1),
				('b', 1),
			])
		])

		self.assertFieldEqual(layout['cyc'], (unsigned(1), DIR_NONE))
		self.assertFieldEqual(layout['data'], (signed(32), DIR_NONE))
		self.assertFieldEqual(layout['stb'], (unsigned(1), DIR_FANOUT))
		self.assertFieldEqual(layout['ack'], (unsigned(1), DIR_FANIN))
		sublayout = layout['info'][0]
		self.assertEqual(layout['info'][1], DIR_NONE)
		self.assertFieldEqual(sublayout['a'], (unsigned(1), DIR_NONE))
		self.assertFieldEqual(sublayout['b'], (unsigned(1), DIR_NONE))

	def test_enum_field(self):
		layout = Layout.cast([
			('enum', UnsignedEnum),
			('enum_dir', UnsignedEnum, DIR_FANOUT),
		])
		self.assertFieldEqual(layout['enum'], (unsigned(2), DIR_NONE))
		self.assertFieldEqual(layout['enum_dir'], (unsigned(2), DIR_FANOUT))

	def test_range_field(self):
		layout = Layout.cast([
			('range', range(0, 7)),
		])
		self.assertFieldEqual(layout['range'], (unsigned(3), DIR_NONE))

	def test_slice_tuple(self):
		layout = Layout.cast([
			('a', 1),
			('b', 2),
			('c', 3)
		])
		expect = Layout.cast([
			('a', 1),
			('c', 3)
		])
		self.assertEqual(layout['a', 'c'], expect)

	def test_repr(self):
		self.assertEqual(
			repr(Layout([('a', unsigned(1)), ('b', signed(2))])),
			'Layout([(\'a\', unsigned(1)), (\'b\', signed(2))])'
		)
		self.assertEqual(
			repr(Layout([('a', unsigned(1)), ('b', [('c', signed(3))])])),
			'Layout([(\'a\', unsigned(1)), '
			'(\'b\', Layout([(\'c\', signed(3))]))])'
		)

	def test_wrong_field(self):
		with self.assertRaisesRegex(
			TypeError, (
				r'^Field \(1,\) has invalid layout: should be either \(name, shape\) or '
				r'\(name, shape, direction\)$'
			)
		):
			Layout.cast([(1,)])

	def test_wrong_name(self):
		with self.assertRaisesRegex(
			TypeError,
			r'^Field \(1, 1\) has invalid name: should be a string$'
		):
			Layout.cast([(1, 1)])

	def test_wrong_name_duplicate(self):
		with self.assertRaisesRegex(
			NameError,
			r'^Field \(\'a\', 2\) has a name that is already present in the layout$'
		):
			Layout.cast([('a', 1), ('a', 2)])

	def test_wrong_direction(self):
		with self.assertRaisesRegex(
			TypeError, (
				r'^Field \(\'a\', 1, 0\) has invalid direction: should be a Direction '
				r'instance like DIR_FANIN$'
			)
		):
			Layout.cast([('a', 1, 0)])

	def test_wrong_shape(self):
		with self.assertRaisesRegex(
			TypeError, (
				r'^Field \(\'a\', \'x\'\) has invalid shape: should be castable to Shape or '
				r'a list of fields of a nested record$'
			)
		):
			Layout.cast([('a', 'x')])


class RecordTestCase(ToriiTestSuiteCase):
	def test_basic(self):
		r = Record([
			('stb',  1),
			('data', 32),
			('info', [
				('a', 1),
				('b', 1),
			])
		])

		self.assertEqual(repr(r), '(rec r stb data (rec r__info a b))')
		self.assertEqual(len(r),  35)
		self.assertIsInstance(r.stb, Signal)
		self.assertEqual(r.stb.name, 'r__stb')
		self.assertEqual(r['stb'].name, 'r__stb')

		self.assertTrue(hasattr(r, 'stb'))
		self.assertFalse(hasattr(r, 'xxx'))

	def test_unnamed(self):
		r = [Record([
			('stb', 1)
		])][0]

		self.assertEqual(repr(r), '(rec <unnamed> stb)')
		self.assertEqual(r.stb.name, 'stb')

	def test_iter(self):
		r = Record([
			('data', 4),
			('stb',  1),
		])

		self.assertEqual(repr(r[0]),   '(slice (cat (sig r__data) (sig r__stb)) 0:1)')
		self.assertEqual(repr(r[0:3]), '(slice (cat (sig r__data) (sig r__stb)) 0:3)')

	def test_wrong_field(self):
		r = Record([
			('stb', 1),
			('ack', 1),
		])
		with self.assertRaisesRegex(
			AttributeError,
			r'^Record \'r\' does not have a field \'en\'\. Did you mean one of: stb, ack\?$'
		):
			r['en']
		with self.assertRaisesRegex(
			AttributeError,
			r'^Record \'r\' does not have a field \'en\'\. Did you mean one of: stb, ack\?$'
		):
			r.en

	def test_wrong_field_unnamed(self):
		r = [Record([
			('stb', 1),
			('ack', 1),
		])][0]
		with self.assertRaisesRegex(
			AttributeError,
			r'^Unnamed record does not have a field \'en\'\. Did you mean one of: stb, ack\?$'
		):
			r.en

	def test_construct_with_fields(self):
		ns = Signal(1)
		nr = Record([
			('burst', 1)
		])
		r = Record([
			('stb', 1),
			('info', [
				('burst', 1)
			])
		], fields = {
			'stb':  ns,
			'info': nr
		})
		self.assertIs(r.stb, ns)
		self.assertIs(r.info, nr)

	def test_shape(self):
		r1 = Record([('a', 1), ('b', 2)])
		self.assertEqual(r1.shape(), unsigned(3))

	def test_like(self):
		r1 = Record([('a', 1), ('b', 2)])
		r2 = Record.like(r1)
		self.assertEqual(r1.layout, r2.layout)
		self.assertEqual(r2.name, 'r2')
		r3 = Record.like(r1, name = 'foo')
		self.assertEqual(r3.name, 'foo')
		r4 = Record.like(r1, name_suffix = 'foo')
		self.assertEqual(r4.name, 'r1foo')

	def test_like_modifications(self):
		r1 = Record([('a', 1), ('b', [('s', 1)])])
		self.assertEqual(r1.a.name, 'r1__a')
		self.assertEqual(r1.b.name, 'r1__b')
		self.assertEqual(r1.b.s.name, 'r1__b__s')
		r1.a.reset = 1
		r1.b.s.reset = 1
		r2 = Record.like(r1)
		self.assertEqual(r2.a.reset, 1)
		self.assertEqual(r2.b.s.reset, 1)
		self.assertEqual(r2.a.name, 'r2__a')
		self.assertEqual(r2.b.name, 'r2__b')
		self.assertEqual(r2.b.s.name, 'r2__b__s')

	def test_slice_tuple(self):
		r1 = Record([('a', 1), ('b', 2), ('c', 3)])
		r2 = r1['a', 'c']
		self.assertEqual(r2.layout, Layout([('a', 1), ('c', 3)]))
		self.assertIs(r2.a, r1.a)
		self.assertIs(r2.c, r1.c)

	def test_enum_decoder(self):
		r1 = Record([('a', UnsignedEnum)])
		self.assertEqual(r1.a.decoder(UnsignedEnum.FOO), 'FOO/1')

	def test_operators(self):
		r1 = Record([('a', 1)])
		s1 = Signal()

		# __bool__
		with self.assertRaisesRegex(
			TypeError,
			r'^Attempted to convert Torii value to Python boolean$'
		):
			not r1

		# __invert__, __neg__
		self.assertEqual(repr(~r1), '(~ (cat (sig r1__a)))')
		self.assertEqual(repr(-r1), '(- (cat (sig r1__a)))')

		# __add__, __radd__, __sub__, __rsub__
		self.assertEqual(repr(r1 + 1),  '(+ (cat (sig r1__a)) (const 1\'d1))')
		self.assertEqual(repr(r1 + s1), '(+ (cat (sig r1__a)) (sig s1))')
		self.assertEqual(repr(1 + r1),  '(+ (const 1\'d1) (cat (sig r1__a)))')
		self.assertEqual(repr(s1 + r1), '(+ (sig s1) (cat (sig r1__a)))')
		self.assertEqual(repr(r1 - 1),  '(- (cat (sig r1__a)) (const 1\'d1))')
		self.assertEqual(repr(r1 - s1), '(- (cat (sig r1__a)) (sig s1))')
		self.assertEqual(repr(1 - r1),  '(- (const 1\'d1) (cat (sig r1__a)))')
		self.assertEqual(repr(s1 - r1), '(- (sig s1) (cat (sig r1__a)))')

		# __mul__, __rmul__
		self.assertEqual(repr(r1 * 1),  '(* (cat (sig r1__a)) (const 1\'d1))')
		self.assertEqual(repr(r1 * s1), '(* (cat (sig r1__a)) (sig s1))')
		self.assertEqual(repr(1 * r1),  '(* (const 1\'d1) (cat (sig r1__a)))')
		self.assertEqual(repr(s1 * r1), '(* (sig s1) (cat (sig r1__a)))')

		# __mod__, __rmod__, __floordiv__, __rfloordiv__
		self.assertEqual(repr(r1 % 1),   '(% (cat (sig r1__a)) (const 1\'d1))')
		self.assertEqual(repr(r1 % s1),  '(% (cat (sig r1__a)) (sig s1))')
		self.assertEqual(repr(1 % r1),   '(% (const 1\'d1) (cat (sig r1__a)))')
		self.assertEqual(repr(s1 % r1),  '(% (sig s1) (cat (sig r1__a)))')
		self.assertEqual(repr(r1 // 1),  '(// (cat (sig r1__a)) (const 1\'d1))')
		self.assertEqual(repr(r1 // s1), '(// (cat (sig r1__a)) (sig s1))')
		self.assertEqual(repr(1 // r1),  '(// (const 1\'d1) (cat (sig r1__a)))')
		self.assertEqual(repr(s1 // r1), '(// (sig s1) (cat (sig r1__a)))')

		# __lshift__, __rlshift__, __rshift__, __rrshift__
		self.assertEqual(repr(r1 >> 1),  '(>> (cat (sig r1__a)) (const 1\'d1))')
		self.assertEqual(repr(r1 >> s1), '(>> (cat (sig r1__a)) (sig s1))')
		self.assertEqual(repr(1 >> r1),  '(>> (const 1\'d1) (cat (sig r1__a)))')
		self.assertEqual(repr(s1 >> r1), '(>> (sig s1) (cat (sig r1__a)))')
		self.assertEqual(repr(r1 << 1),  '(<< (cat (sig r1__a)) (const 1\'d1))')
		self.assertEqual(repr(r1 << s1), '(<< (cat (sig r1__a)) (sig s1))')
		self.assertEqual(repr(1 << r1),  '(<< (const 1\'d1) (cat (sig r1__a)))')
		self.assertEqual(repr(s1 << r1), '(<< (sig s1) (cat (sig r1__a)))')

		# __and__, __rand__, __xor__, __rxor__, __or__, __ror__
		self.assertEqual(repr(r1 & 1),  '(& (cat (sig r1__a)) (const 1\'d1))')
		self.assertEqual(repr(r1 & s1), '(& (cat (sig r1__a)) (sig s1))')
		self.assertEqual(repr(1 & r1),  '(& (const 1\'d1) (cat (sig r1__a)))')
		self.assertEqual(repr(s1 & r1), '(& (sig s1) (cat (sig r1__a)))')
		self.assertEqual(repr(r1 ^ 1),  '(^ (cat (sig r1__a)) (const 1\'d1))')
		self.assertEqual(repr(r1 ^ s1), '(^ (cat (sig r1__a)) (sig s1))')
		self.assertEqual(repr(1 ^ r1),  '(^ (const 1\'d1) (cat (sig r1__a)))')
		self.assertEqual(repr(s1 ^ r1), '(^ (sig s1) (cat (sig r1__a)))')
		self.assertEqual(repr(r1 | 1),  '(| (cat (sig r1__a)) (const 1\'d1))')
		self.assertEqual(repr(r1 | s1), '(| (cat (sig r1__a)) (sig s1))')
		self.assertEqual(repr(1 | r1),  '(| (const 1\'d1) (cat (sig r1__a)))')
		self.assertEqual(repr(s1 | r1), '(| (sig s1) (cat (sig r1__a)))')

		# __eq__, __ne__, __lt__, __le__, __gt__, __ge__
		self.assertEqual(repr(r1 == 1),  '(== (cat (sig r1__a)) (const 1\'d1))')
		self.assertEqual(repr(r1 == s1), '(== (cat (sig r1__a)) (sig s1))')
		self.assertEqual(repr(s1 == r1), '(== (sig s1) (cat (sig r1__a)))')
		self.assertEqual(repr(r1 != 1),  '(!= (cat (sig r1__a)) (const 1\'d1))')
		self.assertEqual(repr(r1 != s1), '(!= (cat (sig r1__a)) (sig s1))')
		self.assertEqual(repr(s1 != r1), '(!= (sig s1) (cat (sig r1__a)))')
		self.assertEqual(repr(r1 < 1),   '(< (cat (sig r1__a)) (const 1\'d1))')
		self.assertEqual(repr(r1 < s1),  '(< (cat (sig r1__a)) (sig s1))')
		self.assertEqual(repr(s1 < r1),  '(< (sig s1) (cat (sig r1__a)))')
		self.assertEqual(repr(r1 <= 1),  '(<= (cat (sig r1__a)) (const 1\'d1))')
		self.assertEqual(repr(r1 <= s1), '(<= (cat (sig r1__a)) (sig s1))')
		self.assertEqual(repr(s1 <= r1), '(<= (sig s1) (cat (sig r1__a)))')
		self.assertEqual(repr(r1 > 1),   '(> (cat (sig r1__a)) (const 1\'d1))')
		self.assertEqual(repr(r1 > s1),  '(> (cat (sig r1__a)) (sig s1))')
		self.assertEqual(repr(s1 > r1),  '(> (sig s1) (cat (sig r1__a)))')
		self.assertEqual(repr(r1 >= 1),  '(>= (cat (sig r1__a)) (const 1\'d1))')
		self.assertEqual(repr(r1 >= s1), '(>= (cat (sig r1__a)) (sig s1))')
		self.assertEqual(repr(s1 >= r1), '(>= (sig s1) (cat (sig r1__a)))')

		# __abs__, __len__
		self.assertEqual(repr(abs(r1)), '(cat (sig r1__a))')
		self.assertEqual(len(r1), 1)

		# as_unsigned, as_signed, bool, any, all, xor, implies
		self.assertEqual(repr(r1.as_unsigned()), '(u (cat (sig r1__a)))')
		self.assertEqual(repr(r1.as_signed()),   '(s (cat (sig r1__a)))')
		self.assertEqual(repr(r1.bool()),        '(b (cat (sig r1__a)))')
		self.assertEqual(repr(r1.any()),         '(r| (cat (sig r1__a)))')
		self.assertEqual(repr(r1.all()),         '(r& (cat (sig r1__a)))')
		self.assertEqual(repr(r1.xor()),         '(r^ (cat (sig r1__a)))')
		self.assertEqual(repr(r1.implies(1)),    '(| (~ (cat (sig r1__a))) (const 1\'d1))')
		self.assertEqual(repr(r1.implies(s1)),   '(| (~ (cat (sig r1__a))) (sig s1))')

		# bit_select, word_select, matches,
		self.assertEqual(repr(r1.bit_select(0, 1)),  '(slice (cat (sig r1__a)) 0:1)')
		self.assertEqual(repr(r1.word_select(0, 1)), '(slice (cat (sig r1__a)) 0:1)')
		self.assertEqual(
			repr(r1.matches('1')),
			'(== (& (cat (sig r1__a)) (const 1\'d1)) (const 1\'d1))'
		)

		# shift_left, shift_right, rotate_left, rotate_right, eq
		self.assertEqual(repr(r1.shift_left(1)),  '(cat (const 1\'d0) (cat (sig r1__a)))')
		self.assertEqual(repr(r1.shift_right(1)), '(slice (cat (sig r1__a)) 1:1)')
		self.assertEqual(
			repr(r1.rotate_left(1)),
			'(cat (slice (cat (sig r1__a)) 0:1) (slice (cat (sig r1__a)) 0:0))'
		)
		self.assertEqual(
			repr(r1.rotate_right(1)),
			'(cat (slice (cat (sig r1__a)) 0:1) (slice (cat (sig r1__a)) 0:0))'
		)
		self.assertEqual(repr(r1.eq(1)), '(eq (cat (sig r1__a)) (const 1\'d1))')
		self.assertEqual(repr(r1.eq(s1)), '(eq (cat (sig r1__a)) (sig s1))')


class ConnectTestCase(ToriiTestSuiteCase):
	def setUp_flat(self):
		self.core_layout = [
			('addr',   32, DIR_FANOUT),
			('data_r', 32, DIR_FANIN),
			('data_w', 32, DIR_FANIN),
		]
		self.periph_layout = [
			('addr',   32, DIR_FANOUT),
			('data_r', 32, DIR_FANIN),
			('data_w', 32, DIR_FANIN),
		]

	def setUp_nested(self):
		self.core_layout = [
			('addr',   32, DIR_FANOUT),
			('data', [
				('r',  32, DIR_FANIN),
				('w',  32, DIR_FANIN),
			]),
		]
		self.periph_layout = [
			('addr',   32, DIR_FANOUT),
			('data', [
				('r',  32, DIR_FANIN),
				('w',  32, DIR_FANIN),
			]),
		]

	def test_flat(self):
		self.setUp_flat()

		core    = Record(self.core_layout)
		periph1 = Record(self.periph_layout)
		periph2 = Record(self.periph_layout)

		stmts = core.connect(periph1, periph2)
		self.assertRepr(stmts, '''(
			(eq (sig periph1__addr) (sig core__addr))
			(eq (sig periph2__addr) (sig core__addr))
			(eq (sig core__data_r) (| (sig periph1__data_r) (sig periph2__data_r)))
			(eq (sig core__data_w) (| (sig periph1__data_w) (sig periph2__data_w)))
		)''')

	def test_flat_include(self):
		self.setUp_flat()

		core    = Record(self.core_layout)
		periph1 = Record(self.periph_layout)
		periph2 = Record(self.periph_layout)

		stmts = core.connect(periph1, periph2, include = {'addr': True})
		self.assertRepr(stmts, '''(
			(eq (sig periph1__addr) (sig core__addr))
			(eq (sig periph2__addr) (sig core__addr))
		)''')

	def test_flat_exclude(self):
		self.setUp_flat()

		core    = Record(self.core_layout)
		periph1 = Record(self.periph_layout)
		periph2 = Record(self.periph_layout)

		stmts = core.connect(periph1, periph2, exclude = {'addr': True})
		self.assertRepr(stmts, '''(
			(eq (sig core__data_r) (| (sig periph1__data_r) (sig periph2__data_r)))
			(eq (sig core__data_w) (| (sig periph1__data_w) (sig periph2__data_w)))
		)''')

	def test_nested(self):
		self.setUp_nested()

		core    = Record(self.core_layout)
		periph1 = Record(self.periph_layout)
		periph2 = Record(self.periph_layout)

		stmts = core.connect(periph1, periph2)
		self.maxDiff = None
		self.assertRepr(stmts, '''(
			(eq (sig periph1__addr) (sig core__addr))
			(eq (sig periph2__addr) (sig core__addr))
			(eq (sig core__data__r) (| (sig periph1__data__r) (sig periph2__data__r)))
			(eq (sig core__data__w) (| (sig periph1__data__w) (sig periph2__data__w)))
		)''')

	def test_wrong_include_exclude(self):
		self.setUp_flat()

		core   = Record(self.core_layout)
		periph = Record(self.periph_layout)

		with self.assertRaisesRegex(
			AttributeError,
			r'^Cannot include field \'foo\' because it is not present in record \'core\'$'
		):
			core.connect(periph, include = {'foo': True})

		with self.assertRaisesRegex(
			AttributeError,
			r'^Cannot exclude field \'foo\' because it is not present in record \'core\'$'
		):
			core.connect(periph, exclude = {'foo': True})

	def test_wrong_direction(self):
		recs = [Record([('x', 1)]) for _ in range(2)]

		with self.assertRaisesRegex(
			TypeError, (
				r'^Cannot connect field \'x\' of unnamed record because it does not have '
				r'a direction$'
			)
		):
			recs[0].connect(recs[1])

	def test_wrong_missing_field(self):
		core   = Record([('addr', 32, DIR_FANOUT)])
		periph = Record([])

		with self.assertRaisesRegex(
			AttributeError, (
				r'^Cannot connect field \'addr\' of record \'core\' to subordinate record \'periph\' '
				r'because the subordinate record does not have this field$'
			)
		):
			core.connect(periph)



# SPDX-License-Identifier: BSD-2-Clause

from os            import environ
from shutil        import which

from ..util.string import tool_env_var

__doc__ = '''\

.. warning:

	The following module is internal implementation details for Torii and should
	not be considered stable or for external use.

'''

__all__ = (
	'has_tool',
	'require_tool',
	'ToolNotFound',
)

class ToolNotFound(Exception):
	pass

def _get_tool(name: str) -> str:
	'''
	Get the tool name or path if overridden by an environment variable.

	Parameters
	----------
	name : str
		The tool/executable name.

	Returns
	-------
	str
		The tool/executable name if not found in the environment.

	'''

	return environ.get(tool_env_var(name), name)

def has_tool(name: str) -> bool:
	'''
	Check to see if the tool or executable is in the current PATH.

	Parameters
	----------
	name : str
		The tool/executable name.

	Returns
	-------
	bool
		If the tool/executable was found in the current system PATH.

	'''

	return which(_get_tool(name)) is not None

def require_tool(name: str) -> str:
	'''
	Return the fully resolved path to the requested tool or executable
	if found, otherwise throw a ToolNotFound exception.

	Parameters
	----------
	name : str
		The tool/executable name.

	Returns
	-------
	Path
		The fully qualified path to the executable / tool.

	Raises
	------
	ToolNotFound

	'''
	env_var = tool_env_var(name)
	path = _get_tool(name)
	if which(path) is None:
		if env_var in environ:
			raise ToolNotFound(
				f'Could not find required tool {name} in {path} as specified via '
				f'the {env_var} environment variable'
			)
		else:
			raise ToolNotFound(
				f'Could not find required tool {name} in PATH. Place '
				'it directly in PATH or specify path explicitly '
				f'via the {env_var} environment variable'
			)
	return path



# SPDX-License-Identifier: BSD-2-Clause

from sys    import _getframe, version_info
from typing import Optional, Union

from opcode import opname

__all__ = (
	'get_src_loc',
	'get_var_name',
	'NameNotFound',
)

class NameNotFound(Exception):
	pass

_raise_exception = object()

def get_var_name(depth: int = 2, default: Optional[Union[str, object]] = _raise_exception) -> Union[str, object]:
	frame = _getframe(depth)
	code = frame.f_code
	call_index = frame.f_lasti
	while call_index > 0 and opname[code.co_code[call_index]] == 'CACHE':
		call_index -= 2
	while True:
		call_opc = opname[code.co_code[call_index]]
		if call_opc in ('EXTENDED_ARG',):
			call_index += 2
		else:
			break
	if call_opc not in ('CALL_FUNCTION', 'CALL_FUNCTION_KW', 'CALL_FUNCTION_EX', 'CALL_METHOD', 'CALL'):
		if default is _raise_exception:
			raise NameNotFound
		else:
			return default

	index = call_index + 2
	imm = 0
	while True:
		opc = opname[code.co_code[index]]
		if opc == 'EXTENDED_ARG':
			imm |= int(code.co_code[index + 1])
			imm <<= 8
			index += 2
		elif opc in ('STORE_NAME', 'STORE_ATTR'):
			imm |= int(code.co_code[index + 1])
			return code.co_names[imm]
		elif opc == 'STORE_FAST':
			imm |= int(code.co_code[index + 1])
			if version_info >= (3, 11):
				return code._varname_from_oparg(imm)
			else:
				return code.co_varnames[imm]
		elif opc == 'STORE_DEREF':
			imm |= int(code.co_code[index + 1])
			if version_info >= (3, 11):
				return code._varname_from_oparg(imm)
			else:
				if imm < len(code.co_cellvars):
					return code.co_cellvars[imm]
				else:
					return code.co_freevars[imm - len(code.co_cellvars)]
		elif opc in (
			'LOAD_GLOBAL', 'LOAD_NAME', 'LOAD_ATTR', 'LOAD_FAST',
			'LOAD_DEREF', 'DUP_TOP', 'BUILD_LIST', 'CACHE', 'COPY'
		):
			imm = 0
			index += 2
		else:
			if default is _raise_exception:
				raise NameNotFound
			else:
				return default


def get_src_loc(src_loc_at: int = 0) -> tuple[str, int]:
	# n-th  frame: get_src_loc()
	# n-1th frame: caller of get_src_loc() (usually constructor)
	# n-2th frame: caller of caller (usually user code)
	frame = _getframe(2 + src_loc_at)
	return (frame.f_code.co_filename, frame.f_lineno)



# SPDX-License-Identifier: BSD-2-Clause

from abc         import abstractmethod
from typing      import Literal

from ....build   import Attrs, Clock, Subsignal, TemplatedPlatform
from ....hdl     import (
	ClockDomain, ClockSignal, Const, Instance, Module, Record, ResetSignal, Signal
)
from ....lib.cdc import ResetSynchronizer
from ....lib.io  import Pin

__all__ = (
	'ICE40Platform',
)


class ICE40Platform(TemplatedPlatform):
	'''
	.. rubric:: IceStorm toolchain

	Required tools:
		* ``yosys``
		* ``nextpnr-ice40``
		* ``icepack``

	The environment is populated by running the script specified in the environment variable
	``TORII_ENV_ICESTORM``, if present.

	Available overrides:
		* ``verbose``: enables logging of informational messages to standard error.
		* ``read_verilog_opts``: adds options for ``read_verilog`` Yosys command.
		* ``synth_opts``: adds options for ``synth_ice40`` Yosys command.
		* ``script_after_read``: inserts commands after ``read_rtlil`` in Yosys script.
		* ``script_after_synth``: inserts commands after ``synth_ice40`` in Yosys script.
		* ``yosys_opts``: adds extra options for ``yosys``.
		* ``nextpnr_opts``: adds extra options for ``nextpnr-ice40``.
		* ``add_pre_pack``: inserts commands at the end in pre-pack Python script.
		* ``add_constraints``: inserts commands at the end in the PCF file.

	Build products:
		* ``{{name}}.rpt``: Yosys log.
		* ``{{name}}.json``: synthesized RTL.
		* ``{{name}}.tim``: nextpnr log.
		* ``{{name}}.asc``: ASCII bitstream.
		* ``{{name}}.bin``: binary bitstream.

	.. rubric:: iCECube2 toolchain

	This toolchain comes in two variants: ``LSE-iCECube2`` and ``Synplify-iCECube2``.

	Required tools:
		* iCECube2 toolchain
		* ``tclsh``

	The environment is populated by setting the necessary environment variables based on
	``TORII_ENV_ICECUBE2``, which must point to the root of the iCECube2 installation, and
	is required.

	Available overrides:
		* ``verbose``: enables logging of informational messages to standard error.
		* ``lse_opts``: adds options for LSE.
		* ``script_after_add``: inserts commands after ``add_file`` in Synplify Tcl script.
		* ``script_after_options``: inserts commands after ``set_option`` in Synplify Tcl script.
		* ``add_constraints``: inserts commands in SDC file.
		* ``script_after_flow``: inserts commands after ``run_sbt_backend_auto`` in SBT
		  Tcl script.

	Build products:
		* ``{{name}}_lse.log`` (LSE) or ``{{name}}_design/{{name}}.htm`` (Synplify): synthesis log.
		* ``sbt/outputs/router/{{name}}_timing.rpt``: timing report.
		* ``{{name}}.edf``: EDIF netlist.
		* ``{{name}}.bin``: binary bitstream.

	''' # noqa: E101

	toolchain = None # selected when creating platform

	@property
	@abstractmethod
	def device(self) -> str:
		raise NotImplementedError('Platform must implement this property')

	@property
	@abstractmethod
	def package(self) -> str:
		raise NotImplementedError('Platform must implement this property')

	# IceStorm templates

	_nextpnr_device_options = {
		'iCE40LP384': '--lp384',
		'iCE40LP1K':  '--lp1k',
		'iCE40LP4K':  '--lp8k',
		'iCE40LP8K':  '--lp8k',
		'iCE40HX1K':  '--hx1k',
		'iCE40HX4K':  '--hx8k',
		'iCE40HX8K':  '--hx8k',
		'iCE40UP5K':  '--up5k',
		'iCE40UP3K':  '--up5k',
		'iCE5LP4K':   '--u4k',
		'iCE5LP2K':   '--u4k',
		'iCE5LP1K':   '--u4k',
	}
	_nextpnr_package_options = {
		'iCE40LP4K':  ':4k',
		'iCE40HX4K':  ':4k',
		'iCE40UP3K':  '',
		'iCE5LP2K':   '',
		'iCE5LP1K':   '',
	}

	_icestorm_required_tools = [
		'yosys',
		'nextpnr-ice40',
		'icepack',
	]
	_icestorm_file_templates = {
		**TemplatedPlatform.build_script_templates,
		'{{name}}.il': r'''
			# {{autogenerated}}
			{{emit_rtlil()}}
		''',
		'{{name}}.debug.v': r'''
			/* {{autogenerated}} */
			{{emit_debug_verilog()}}
		''',
		'{{name}}.ys': r'''
			# {{autogenerated}}
			{% for file in platform.iter_files(".v") -%}
				read_verilog {{get_override("read_verilog_opts")|options}} {{file}}
			{% endfor %}
			{% for file in platform.iter_files(".sv") -%}
				read_verilog -sv {{get_override("read_verilog_opts")|options}} {{file}}
			{% endfor %}
			{% for file in platform.iter_files(".il") -%}
				read_rtlil {{file}}
			{% endfor %}
			read_rtlil {{name}}.il
			delete w:$verilog_initial_trigger
			{{get_override("script_after_read")|default("# (script_after_read placeholder)")}}
			synth_ice40 {{get_override("synth_opts")|options}} -top {{name}}
			{{get_override("script_after_synth")|default("# (script_after_synth placeholder)")}}
			write_json {{name}}.json
		''',
		'{{name}}.pcf': r'''
			# {{autogenerated}}
			{% for port_name, pin_name, attrs in platform.iter_port_constraints_bits() -%}
				set_io {{port_name}} {{pin_name}}
			{% endfor %}
			{% for net_signal, port_signal, frequency in platform.iter_clock_constraints() -%}
				set_frequency {{net_signal|hierarchy(".")}} {{frequency/1000000}}
			{% endfor%}
			{{get_override("add_constraints")|default("# (add_constraints placeholder)")}}
		''',
	}
	_icestorm_command_templates = [
		r'''
		{{invoke_tool("yosys")}}
			{{quiet("-q")}}
			{{get_override("yosys_opts")|options}}
			-l {{name}}.rpt
			{{name}}.ys
		''',
		r'''
		{{invoke_tool("nextpnr-ice40")}}
			{{quiet("--quiet")}}
			{{get_override("nextpnr_opts")|options}}
			--log {{name}}.tim
			{{platform._nextpnr_device_options[platform.device]}}
			--package
				{{platform.package|lower}}{{platform._nextpnr_package_options[platform.device]|
											default("")}}
			--json {{name}}.json
			--pcf {{name}}.pcf
			--asc {{name}}.asc
		''',
		r'''
		{{invoke_tool("icepack")}}
			{{verbose("-v")}}
			{{get_override("icepack_opts")|options}}
			{{name}}.asc
			{{name}}.bin
		'''
	]

	# iCECube2 templates

	_icecube2_required_tools = [
		'synthesis',
		'synpwrap',
		'tclsh',
	]
	_icecube2_file_templates = {
		**TemplatedPlatform.build_script_templates,
		'build_{{name}}.sh': r'''
			# {{autogenerated}}
			set -e{{verbose("x")}}
			if [ -n "${{platform._toolchain_env_var}}" ]; then
				# LSE environment
				export LD_LIBRARY_PATH=${{platform._toolchain_env_var}}/LSE/bin/lin64:$LD_LIBRARY_PATH
				export PATH=${{platform._toolchain_env_var}}/LSE/bin/lin64:$PATH
				export FOUNDRY=${{platform._toolchain_env_var}}/LSE
				# Synplify environment
				export LD_LIBRARY_PATH=${{platform._toolchain_env_var}}/sbt_backend/bin/linux/opt/synpwrap:$LD_LIBRARY_PATH
				export PATH=${{platform._toolchain_env_var}}/sbt_backend/bin/linux/opt/synpwrap:$PATH
				export SYNPLIFY_PATH=${{platform._toolchain_env_var}}/synpbase
				# Common environment
				export SBT_DIR=${{platform._toolchain_env_var}}/sbt_backend
			else
				echo "Variable ${{platform._toolchain_env_var}} must be set" >&2; exit 1
			fi
			{{emit_commands("sh")}}
		''',
		'{{name}}.v': r'''
			/* {{autogenerated}} */
			{{emit_verilog()}}
		''',
		'{{name}}.debug.v': r'''
			/* {{autogenerated}} */
			{{emit_debug_verilog()}}
		''',
		'{{name}}_lse.prj': r'''
			# {{autogenerated}}
			-a SBT{{platform.family}}
			-d {{platform.device}}
			-t {{platform.package}}
			{{get_override("lse_opts")|options|default("# (lse_opts placeholder)")}}
			{% for file in platform.iter_files(".v") -%}
				-ver {{file}}
			{% endfor %}
			-ver {{name}}.v
			-sdc {{name}}.sdc
			-top {{name}}
			-output_edif {{name}}.edf
			-logfile {{name}}_lse.log
		''',
		'{{name}}_syn.prj': r'''
			# {{autogenerated}}
			{% for file in platform.iter_files(".v", ".sv", ".vhd", ".vhdl") -%}
				add_file -verilog {{file|tcl_escape}}
			{% endfor %}
			add_file -verilog {{name}}.v
			add_file -constraint {{name}}.sdc
			{{get_override("script_after_add")|default("# (script_after_add placeholder)")}}
			impl -add {{name}}_design -type fpga
			set_option -technology SBT{{platform.family}}
			set_option -part {{platform.device}}
			set_option -package {{platform.package}}
			{{get_override("script_after_options")|default("# (script_after_options placeholder)")}}
			project -result_format edif
			project -result_file {{name}}.edf
			impl -active {{name}}_design
			project -run compile
			project -run map
			project -run fpga_mapper
			file copy -force -- {{name}}_design/{{name}}.edf {{name}}.edf
		''',
		'{{name}}.sdc': r'''
			# {{autogenerated}}
			{% for net_signal, port_signal, frequency in platform.iter_clock_constraints() -%}
				{% if port_signal is not none -%}
					create_clock -name {{port_signal.name|tcl_escape}} -period {{1000000000/frequency}} [get_ports {{port_signal.name|tcl_escape}}]
				{% else -%}
					create_clock -name {{net_signal.name|tcl_escape}} -period {{1000000000/frequency}} [get_nets {{net_signal|hierarchy("/")|tcl_escape}}]
				{% endif %}
			{% endfor %}
			{{get_override("add_constraints")|default("# (add_constraints placeholder)")}}
		''', # noqa: E501
		'{{name}}.tcl': r'''
			# {{autogenerated}}
			set device {{platform.device}}-{{platform.package}}
			set top_module {{name}}
			set proj_dir .
			set output_dir .
			set edif_file {{name}}
			set tool_options ":edifparser -y {{name}}.pcf"
			set sbt_root $::env(SBT_DIR)
			append sbt_tcl $sbt_root "/tcl/sbt_backend_synpl.tcl"
			source $sbt_tcl
			run_sbt_backend_auto $device $top_module $proj_dir $output_dir $tool_options $edif_file
			{{get_override("script_after_file")|default("# (script_after_file placeholder)")}}
			file copy -force -- sbt/outputs/bitmap/{{name}}_bitmap.bin {{name}}.bin
			exit
		''',
		'{{name}}.pcf': r'''
			# {{autogenerated}}
			{% for port_name, pin_name, attrs in platform.iter_port_constraints_bits() -%}
				set_io {{port_name}} {{pin_name}}
			{% endfor %}
		''',
	}
	_lse_icecube2_command_templates = [
		r'''synthesis -f {{name}}_lse.prj''',
		r'''tclsh {{name}}.tcl''',
	]
	_synplify_icecube2_command_templates = [
		r'''synpwrap -prj {{name}}_syn.prj -log {{name}}_syn.log''',
		r'''tclsh {{name}}.tcl''',
	]

	# Common logic

	def __init__(
		self, *, toolchain: Literal['IceStorm', 'LSE-iCECube2', 'Synplify-iCECube2'] = 'IceStorm'
	) -> None:
		super().__init__()

		assert toolchain in ('IceStorm', 'LSE-iCECube2', 'Synplify-iCECube2')
		self.toolchain = toolchain

	@property
	def family(self) -> str:
		if self.device.startswith('iCE40'):
			return 'iCE40'
		if self.device.startswith('iCE5'):
			return 'iCE5'
		assert False

	@property
	def _toolchain_env_var(self) -> str:
		if self.toolchain in ('LSE-iCECube2', 'Synplify-iCECube2'):
			return 'TORII_ENV_iCECube2'
		elif self.toolchain == 'IceStorm':
			return super()._toolchain_env_var
		assert False

	@property
	def required_tools(self) -> list[str]:
		if self.toolchain == 'IceStorm':
			return self._icestorm_required_tools
		if self.toolchain in ('LSE-iCECube2', 'Synplify-iCECube2'):
			return self._icecube2_required_tools
		assert False

	@property
	def file_templates(self) -> dict[str, str]:
		if self.toolchain == 'IceStorm':
			return self._icestorm_file_templates
		if self.toolchain in ('LSE-iCECube2', 'Synplify-iCECube2'):
			return self._icecube2_file_templates
		assert False

	@property
	def command_templates(self) -> list[str]:
		if self.toolchain == 'IceStorm':
			return self._icestorm_command_templates
		if self.toolchain == 'LSE-iCECube2':
			return self._lse_icecube2_command_templates
		if self.toolchain == 'Synplify-iCECube2':
			return self._synplify_icecube2_command_templates
		assert False

	@property
	def default_clk_constraint(self) -> Clock:
		# Internal high-speed oscillator: 48 MHz / (2 ^ div)
		if self.default_clk == 'SB_HFOSC':
			return Clock(48e6 / 2 ** self.hfosc_div)
		# Internal low-speed oscillator: 10 KHz
		elif self.default_clk == 'SB_LFOSC':
			return Clock(10e3)
		# Otherwise, use the defined Clock resource.
		return super().default_clk_constraint

	def create_missing_domain(self, name: str) -> Module:
		# For unknown reasons (no errata was ever published, and no documentation mentions this
		# issue), iCE40 BRAMs read as zeroes for ~3 us after configuration and release of internal
		# global reset. Note that this is a *time-based* delay, generated purely by the internal
		# oscillator, which may not be observed nor influenced directly. For details, see links:
		#  * https://github.com/cliffordwolf/icestorm/issues/76#issuecomment-289270411
		#  * https://github.com/cliffordwolf/icotools/issues/2#issuecomment-299734673
		#
		# To handle this, it is necessary to have a global reset in any iCE40 design that may
		# potentially instantiate BRAMs, and assert this reset for >3 us after configuration.
		# (We add a margin of 5x to allow for PVT variation.) If the board includes a dedicated
		# reset line, this line is ORed with the power on reset.
		#
		# If an internal oscillator is selected as the default clock source, the power-on-reset
		# delay is increased to 100 us, since the oscillators are only stable after that long.
		#
		# The power-on reset timer counts up because the vendor tools do not support initialization
		# of flip-flops.
		if name == 'sync' and self.default_clk is not None:
			m = Module()

			# Internal high-speed clock: 6 MHz, 12 MHz, 24 MHz, or 48 MHz depending on the divider.
			if self.default_clk == 'SB_HFOSC':
				if not hasattr(self, 'hfosc_div'):
					raise ValueError('SB_HFOSC divider exponent (hfosc_div) must be an integer between 0 and 3')
				if not isinstance(self.hfosc_div, int) or self.hfosc_div < 0 or self.hfosc_div > 3:
					raise ValueError(
						f'SB_HFOSC divider exponent (hfosc_div) must be an integer between 0 and 3, not {self.hfosc_div!r}'
					)
				clk_i = Signal()
				m.submodules += Instance(
					'SB_HFOSC',
					i_CLKHFEN = 1,
					i_CLKHFPU = 1,
					p_CLKHF_DIV = f'0b{self.hfosc_div:02b}',
					o_CLKHF = clk_i
				)
				delay = int(100e-6 * self.default_clk_frequency)
			# Internal low-speed clock: 10 KHz.
			elif self.default_clk == 'SB_LFOSC':
				clk_i = Signal()
				m.submodules += Instance(
					'SB_LFOSC',
					i_CLKLFEN = 1,
					i_CLKLFPU = 1,
					o_CLKLF = clk_i
				)
				delay = int(100e-6 * self.default_clk_frequency)
			# User-defined clock signal.
			else:
				clk_i = self.request(self.default_clk).i
				delay = int(15e-6 * self.default_clk_frequency)

			if self.default_rst is not None:
				rst_i = self.request(self.default_rst).i
			else:
				rst_i = Const(0)

			# Power-on-reset domain
			m.domains += ClockDomain('por', reset_less = True, local = True)
			timer = Signal(range(delay))
			ready = Signal()
			m.d.comb += ClockSignal('por').eq(clk_i)
			with m.If(timer == delay):
				m.d.por += ready.eq(1)
			with m.Else():
				m.d.por += timer.eq(timer + 1)

			# Primary domain
			m.domains += ClockDomain('sync')
			m.d.comb += ClockSignal('sync').eq(clk_i)
			if self.default_rst is not None:
				m.submodules.reset_sync = ResetSynchronizer(~ready | rst_i, domain = 'sync')
			else:
				m.d.comb += ResetSignal('sync').eq(~ready)

			return m

	def should_skip_port_component(
		self, port: Subsignal, attrs: Attrs, component: Literal['io', 'i', 'o', 'p', 'n', 'oe']
	) -> bool:
		# On iCE40, a differential input is placed by only instantiating an SB_IO primitive for
		# the pin with z = 0, which is the non-inverting pin. The pinout unfortunately differs
		# between LP/HX and UP series:
		#  * for LP/HX, z = 0 is DPxxB   (B is non-inverting, A is inverting)
		#  * for UP,    z = 0 is IOB_xxA (A is non-inverting, B is inverting)
		if attrs.get('IO_STANDARD', 'SB_LVCMOS') == 'SB_LVDS_INPUT' and component == 'n':
			return True
		return False

	def _get_io_buffer(
		self, m: Module, pin: Pin, port: Subsignal, attrs: Attrs, *,
		i_invert: bool = False, o_invert: bool = False, invert_lut: bool = False
	) -> None:
		def get_dff(clk: Signal, d: Signal, q: Signal) -> None:
			m.submodules += Instance(
				'$dff',
				p_CLK_POLARITY = 1,
				p_WIDTH = len(d),
				i_CLK = clk,
				i_D = d,
				o_Q = q
			)

		def get_ineg(y: Signal, invert: bool) -> Signal:
			if invert_lut:
				a = Signal.like(y, name_suffix = f'_x{1 if invert else 0}')
				for bit in range(len(y)):
					m.submodules += Instance(
						'SB_LUT4',
						p_LUT_INIT = Const(0b01 if invert else 0b10, 16),
						i_I0 = a[bit],
						i_I1 = Const(0),
						i_I2 = Const(0),
						i_I3 = Const(0),
						o_O = y[bit]
					)
				return a
			elif invert:
				a = Signal.like(y, name_suffix = '_n')
				m.d.comb += y.eq(~a)
				return a
			else:
				return y

		def get_oneg(a: Signal, invert: bool) -> Signal:
			if invert_lut:
				y = Signal.like(a, name_suffix = f'_x{1 if invert else 0}')
				for bit in range(len(a)):
					m.submodules += Instance(
						'SB_LUT4',
						p_LUT_INIT = Const(0b01 if invert else 0b10, 16),
						i_I0 = a[bit],
						i_I1 = Const(0),
						i_I2 = Const(0),
						i_I3 = Const(0),
						o_O = y[bit]
					)
				return y
			elif invert:
				y = Signal.like(a, name_suffix = '_n')
				m.d.comb += y.eq(~a)
				return y
			else:
				return a

		if 'GLOBAL' in attrs:
			is_global_input = bool(attrs['GLOBAL'])
			del attrs['GLOBAL']
		else:
			is_global_input = False
		assert not (is_global_input and i_invert)

		if 'i' in pin.dir:
			if pin.xdr < 2:
				pin_i  = get_ineg(pin.i,  i_invert)
			elif pin.xdr == 2:
				pin_i0 = get_ineg(pin.i0, i_invert)
				pin_i1 = get_ineg(pin.i1, i_invert)
		if 'o' in pin.dir:
			if pin.xdr < 2:
				pin_o  = get_oneg(pin.o,  o_invert)
			elif pin.xdr == 2:
				pin_o0 = get_oneg(pin.o0, o_invert)
				pin_o1 = get_oneg(pin.o1, o_invert)

		if 'i' in pin.dir and pin.xdr == 2:
			i0_ff = Signal.like(pin_i0, name_suffix = '_ff')
			i1_ff = Signal.like(pin_i1, name_suffix = '_ff')
			get_dff(pin.i_clk, i0_ff, pin_i0)
			get_dff(pin.i_clk, i1_ff, pin_i1)
		if 'o' in pin.dir and pin.xdr == 2:
			o1_ff = Signal.like(pin_o1, name_suffix = '_ff')
			get_dff(pin.o_clk, pin_o1, o1_ff)

		for bit in range(len(port)):
			io_args = [
				('io', 'PACKAGE_PIN', port[bit]),
				*(('p', key, value) for key, value in attrs.items()),
			]

			if 'i' not in pin.dir:
				# If no input pin is requested, it is important to use a non-registered input pin
				# type, because an output-only pin would not have an input clock, and if its input
				# is configured as registered, this would prevent a co-located input-capable pin
				# from using an input clock.
				i_type = 0b01 # PIN_INPUT
			elif pin.xdr == 0:
				i_type = 0b01 # PIN_INPUT
			elif pin.xdr > 0:
				i_type = 0b00 # PIN_INPUT_REGISTERED aka PIN_INPUT_DDR
			if 'o' not in pin.dir:
				o_type = 0b0000   # PIN_NO_OUTPUT
			elif pin.xdr == 0 and pin.dir == 'o':
				o_type = 0b0110   # PIN_OUTPUT
			elif pin.xdr == 0:
				o_type = 0b1010   # PIN_OUTPUT_TRISTATE
			elif pin.xdr == 1 and pin.dir == 'o':
				o_type = 0b0101   # PIN_OUTPUT_REGISTERED
			elif pin.xdr == 1:
				o_type = 0b1101   # PIN_OUTPUT_REGISTERED_ENABLE_REGISTERED
			elif pin.xdr == 2 and pin.dir == 'o':
				o_type = 0b0100   # PIN_OUTPUT_DDR
			elif pin.xdr == 2:
				o_type = 0b1100   # PIN_OUTPUT_DDR_ENABLE_REGISTERED
			io_args.append(('p', 'PIN_TYPE', Const((o_type << 2) | i_type, 6)))

			if hasattr(pin, 'i_clk'):
				io_args.append(('i', 'INPUT_CLK',  pin.i_clk))
			if hasattr(pin, 'o_clk'):
				io_args.append(('i', 'OUTPUT_CLK', pin.o_clk))

			if 'i' in pin.dir:
				if pin.xdr == 0 and is_global_input:
					io_args.append(('o', 'GLOBAL_BUFFER_OUTPUT', pin.i[bit]))
				elif pin.xdr < 2:
					io_args.append(('o', 'D_IN_0',  pin_i[bit]))
				elif pin.xdr == 2:
					# Re-register both inputs before they enter fabric. This increases hold time
					# to an entire cycle, and adds one cycle of latency.
					io_args.append(('o', 'D_IN_0',  i0_ff[bit]))
					io_args.append(('o', 'D_IN_1',  i1_ff[bit]))
			if 'o' in pin.dir:
				if pin.xdr < 2:
					io_args.append(('i', 'D_OUT_0', pin_o[bit]))
				elif pin.xdr == 2:
					# Re-register neg-edge output after it leaves fabric. This increases setup time
					# to an entire cycle, and doesn't add latency.
					io_args.append(('i', 'D_OUT_0', pin_o0[bit]))
					io_args.append(('i', 'D_OUT_1', o1_ff[bit]))

			if pin.dir in ('oe', 'io'):
				io_args.append(('i', 'OUTPUT_ENABLE', pin.oe))

			if is_global_input:
				m.submodules[f'{pin.name}_{bit}'] = Instance('SB_GB_IO', *io_args)
			else:
				m.submodules[f'{pin.name}_{bit}'] = Instance('SB_IO', *io_args)

	def get_input(self, pin: Pin, port: Record, attrs: Attrs, invert: bool) -> Module:
		self._check_feature(
			'single-ended input', pin, attrs, valid_xdrs = (0, 1, 2), valid_attrs = True
		)

		m = Module()
		self._get_io_buffer(m, pin, port.io, attrs, i_invert = invert)
		return m

	def get_output(self, pin: Pin, port: Record, attrs: Attrs, invert: bool) -> Module:
		self._check_feature(
			'single-ended output', pin, attrs, valid_xdrs = (0, 1, 2), valid_attrs = True
		)

		m = Module()
		self._get_io_buffer(m, pin, port.io, attrs, o_invert = invert)
		return m

	def get_tristate(self, pin: Pin, port: Record, attrs: Attrs, invert: bool) -> Module:
		self._check_feature(
			'single-ended tristate', pin, attrs, valid_xdrs = (0, 1, 2), valid_attrs = True
		)

		m = Module()
		self._get_io_buffer(m, pin, port.io, attrs, o_invert = invert)
		return m

	def get_input_output(self, pin: Pin, port: Record, attrs: Attrs, invert: bool) -> Module:
		self._check_feature(
			'single-ended input/output', pin, attrs, valid_xdrs = (0, 1, 2), valid_attrs = True
		)

		m = Module()
		self._get_io_buffer(m, pin, port.io, attrs, i_invert = invert, o_invert = invert)
		return m

	def get_diff_input(self, pin: Pin, port: Record, attrs: Attrs, invert: bool) -> Module:
		self._check_feature(
			'differential input', pin, attrs, valid_xdrs = (0, 1, 2), valid_attrs = True
		)

		m = Module()
		# See comment in should_skip_port_component above.
		self._get_io_buffer(m, pin, port.p, attrs, i_invert = invert)
		return m

	def get_diff_output(self, pin: Pin, port: Record, attrs: Attrs, invert: bool) -> Module:
		self._check_feature(
			'differential output', pin, attrs, valid_xdrs = (0, 1, 2), valid_attrs = True
		)

		m = Module()
		# Note that the non-inverting output pin is not driven the same way as a regular
		# output pin. The inverter introduces a delay, so for a non-inverting output pin,
		# an identical delay is introduced by instantiating a LUT. This makes the waveform
		# perfectly symmetric in the xdr = 0 case.
		self._get_io_buffer(m, pin, port.p, attrs, o_invert = invert, invert_lut = True)
		self._get_io_buffer(m, pin, port.n, attrs, o_invert = not invert, invert_lut = True)
		return m

	# Tristate bidirectional buffers are not supported on iCE40 because it requires external
	# termination, which is different for differential pins configured as inputs and outputs.

	# CDC primitives are not currently specialized for iCE40. It is not known if iCECube2 supports
	# the necessary attributes; nextpnr-ice40 does not.
